"""
Interface for instruction management.
"""
import logging
from typing import Optional

from kordesii.utils.function_tracing import utils
from kordesii.utils.function_tracing.ARM.operands import ARMOperand
from kordesii.utils.function_tracing.exceptions import FunctionTracingError
from kordesii.utils.function_tracing.instruction import Instruction
from . import ida_arm

logger = logging.getLogger(__name__)


class ARMInstruction(Instruction):

    _operand_class = ARMOperand

    # Maps condition code to function to check if we can execute.
    _cond_map = {
        ida_arm.cond_t.cEQ: lambda ctx: bool(ctx.registers.z),
        ida_arm.cond_t.cNE: lambda ctx: bool(not ctx.registers.z),
        ida_arm.cond_t.cCS: lambda ctx: bool(ctx.registers.c),
        ida_arm.cond_t.cCC: lambda ctx: bool(not ctx.registers.c),
        ida_arm.cond_t.cMI: lambda ctx: bool(ctx.registers.n),
        ida_arm.cond_t.cPL: lambda ctx: bool(not ctx.registers.n),
        ida_arm.cond_t.cVS: lambda ctx: bool(ctx.registers.v),
        ida_arm.cond_t.cVC: lambda ctx: bool(not ctx.registers.v),
        ida_arm.cond_t.cHI: lambda ctx: bool(ctx.registers.c and not ctx.registers.z),
        ida_arm.cond_t.cLS: lambda ctx: bool(not ctx.registers.c or ctx.registers.z),
        ida_arm.cond_t.cGE: lambda ctx: bool(
            (ctx.registers.n and ctx.registers.v)
            or (not ctx.registers.n and not ctx.registers.v)
        ),
        ida_arm.cond_t.cLT: lambda ctx: bool(
            (ctx.registers.n and not ctx.registers.v)
            or (not ctx.registers.n and ctx.registers.v)
        ),
        ida_arm.cond_t.cGT: lambda ctx: bool(
            not ctx.registers.z
            and (
                (ctx.registers.n and ctx.registers.v)
                or (not ctx.registers.n and not ctx.registers.v)
            )
        ),
        ida_arm.cond_t.cLE: lambda ctx: bool(
            ctx.registers.z
            or (ctx.registers.n and not ctx.registers.v)
            or (not ctx.registers.n and ctx.registers.v)
        ),
        ida_arm.cond_t.cAL: lambda ctx: True,
        ida_arm.cond_t.cNV: lambda ctx: False,
    }

    @property
    def flag_update(self) -> bool:
        """
        Whether the condition flags are updated on the result of the operation.
        (S postfix)
        """
        return bool(self._insn.auxpref & ida_arm.aux_cond)

    def _check_condition(self) -> bool:
        """
        Checks condition flags to determine if instruction should be executed.
        """
        condition = ida_arm.get_cond(self._insn)
        try:
            return self._cond_map[condition](self._cpu_context)
        except IndexError:
            raise FunctionTracingError(f"Invalid condition code: {condition}")

    def _execute(self):
        # First check if conditions allow us to execute the instruction.
        if not self._check_condition():
            logger.debug("Skipping instruction at 0x%X. Condition code fails.", self.ip)
            return

        # Execute instruction.
        super()._execute()

        # If post-index or pre-index with update addressing mode,
        # update operand's base register based on offset.
        if self._insn.auxpref & (ida_arm.aux_wback | ida_arm.aux_wbackldm | ida_arm.aux_postidx):
            if self._insn.auxpref & ida_arm.aux_wbackldm:
                # write back for first operand (LDM/STM instruction)
                operand = self.operands[0]
            else:
                # write back for last operand (! postfix or post-indexed operand)
                operand = self.operands[-1]
            logger.debug("Post-index operation for: %s", operand.text)
            operand.base += operand.offset




"""
Interface for operand management.
"""

import collections
import logging
from copy import deepcopy
from typing import Optional

import ida_frame
import ida_nalt
import ida_typeinf
import ida_ua
import idaapi
import idc
import numpy

from kordesii.utils.function_tracing import utils
from kordesii.utils.function_tracing.exceptions import FunctionTracingError
from .x86_64 import ida_intel

logger = logging.getLogger(__name__)


class Operand:
    """Stores information for a given operand for a specific CPU context state."""

    def __init__(self, cpu_context, ip, idx, _type=None):
        """
        :param cpu_context: CPU context to pull operand value
        :param ip: instruction pointer
        :param idx: operand number (0 = first operand, 1 = second operand, ...)
        :param _type: Type of operand
            (Sometimes provided to help avoid recomputation.)
        """
        self.ip = ip
        self.idx = idx

        if _type is not None:
            self.type = _type
        else:
            self.type = idc.get_operand_type(ip, idx)

        self.text = idc.print_operand(ip, idx)
        self._cpu_context = cpu_context
        self._width = None
        self.__insn = None

    def __repr__(self):
        string = f"<{self.__class__.__name__} 0x{self.ip:0x}:{self.idx} : {self.text} = {self.value!r}"
        if self.addr is not None:
            string += f" : &{self.text} = 0x{self.addr:0x}"
        string += f" : width = {self.width}>"
        return string

    def __deepcopy__(self, memo):
        # When we deep copy, clear out the __insn attribute so we don't
        # run into any serialization issues with Swig objects.
        deepcopy_method = self.__deepcopy__
        self.__deepcopy__ = None
        self.__insn = None
        copy = deepcopy(self, memo)
        self.__deepcopy__ = deepcopy_method
        return copy

    @property
    def _tif(self):
        tif = ida_typeinf.tinfo_t()
        ida_nalt.get_op_tinfo(tif, self.ip, self.idx)
        return tif

    @property
    def _insn(self) -> ida_ua.insn_t:
        if self.__insn:
            return self.__insn
        insn = ida_ua.insn_t()
        ida_ua.decode_insn(insn, self.ip)
        if not insn:
            raise FunctionTracingError("Failed to decode instruction at 0x:{:X}".format(self.ip))
        self.__insn = insn
        return self.__insn

    @property
    def _op(self) -> ida_ua.op_t:
        return self._insn.ops[self.idx]

    def _record_stack_variable(self, addr):
        """
        Record the stack variable encountered at the given address.
        """
        # Ignore if base is 0, because that means we don't have enough information to designate this to a variable.
        if self.base:
            stack_var = ida_frame.get_stkvar(self._insn, self._op, self.offset)
            if stack_var:
                frame_id = idc.get_frame_id(self.ip)
                member, stack_offset = stack_var
                # If the offset in the member object is different than the given stack_offset
                # then we are indexing into a variable.
                # We need to adjust the address to be pointing to the base variable address.
                var_addr = addr - (stack_offset - member.soff)
                self._cpu_context.variables.add(
                    var_addr, frame_id=frame_id, stack_offset=member.soff, reference=self.ip
                )

    @property
    def width(self):
        """
        Based on the dtyp value, the size of the operand in bytes

        :return: size of data type
        """
        if self._width is None:
            self._width = ida_ua.get_dtype_size(self._op.dtype)
        return self._width

    @property
    def is_hidden(self):
        """
        True if the operand is not part of the visible assembly code.
        (These are for implicit registers like EAX)
        """
        return self.text == "" or self.is_void

    @property
    def is_void(self):
        """
        True if the operand is not valid.
        """
        return self.type == idc.o_void

    @property
    def is_register(self):
        """True if the operand is a single register."""
        return self.type == idc.o_reg

    @property
    def has_register(self):
        """True if the operand contains a register."""
        return self.type in (idc.o_reg, idc.o_displ, idc.o_phrase, idc.o_fpreg, idc.o_trreg, idc.o_creg, idc.o_dbreg)

    @property
    def is_immediate(self):
        """True if the operand is an immediate value."""
        return self.type in (idc.o_imm, idc.o_near, idc.o_far)

    @property
    def is_memory_reference(self):
        """True if the operand is a memory reference."""
        return self.type in (idc.o_mem, idc.o_phrase, idc.o_displ)

    @property
    def has_phrase(self):
        """True if the operand contains a phrase."""
        return self.type in (idc.o_phrase, idc.o_displ)

    @property
    def is_func_ptr(self):
        """True if the operand is a pointer to a function."""
        return utils.is_func_ptr(self.addr or self.value)

    @property
    def offset(self) -> Optional[int]:
        """The offset value if the operand is a displacement."""
        return None

    @property
    def base(self) -> Optional[int]:
        """The value of the base register if operand is a displacement."""
        return None

    @property
    def addr(self) -> Optional[int]:
        """
        Retrieves the referenced memory address of the operand.

        This should be overwritten by architecture specific Operand implementations
        if this property is applicable.

        :return int: Memory address or None if operand is not a memory reference.
        """
        return None

    @property
    def value(self):
        """
        Retrieve the value of the operand as it is currently in the cpu_context.
        NOTE: We can't cache this value because the value may change based on the cpu context.

        :return int: An integer of the operand value.
        """
        if self.is_hidden:
            return None

        if self.is_immediate:
            value = self._op.value if self.type == idc.o_imm else self._op.addr
            # Create variable/reference if global.
            if idc.is_loaded(value):
                self._cpu_context.variables.add(value, reference=self.ip)
            return value

        if self.is_register:
            value = self._cpu_context.registers[self.text]
            # Record reference if register is a variable address.
            if value in self._cpu_context.variables:
                self._cpu_context.variables[value].add_reference(self.ip)
            return value

        # If a memory reference, return read in memory.
        if self.is_memory_reference:
            addr = self.addr

            # Record reference if address is a variable address.
            if addr in self._cpu_context.variables:
                self._cpu_context.variables[addr].add_reference(self.ip)

            # If a function pointer, we want to return the address.
            # This is because a function may be seen as a memory reference, but we don't
            # want to dereference it in case it in a non-call instruction.
            # (e.g.  "mov  esi, ds:LoadLibraryA")
            # NOTE: Must use internal function to avoid recursive loop.
            if utils.is_func_ptr(addr):
                return addr

            # Return empty
            if not self.width:
                logger.debug("Width is zero for %s, returning empty string.", self.text)
                return b""

            # Otherwise, dereference the address.
            value = self._cpu_context.mem_read(addr, self.width)
            return utils.struct_unpack(value)

        raise FunctionTracingError("Invalid operand type: {}".format(self.type), ip=self.ip)

    @value.setter
    def value(self, value):
        """
        Set the operand to the specified value within the cpu_context.
        """
        # Value may be signed.
        if isinstance(value, int) and value < 0:
            value = utils.unsigned(value, bit_width=self.width * 8)

        # If we are writing to an immediate, I believe they want to write to the memory at the immediate.
        # TODO: Should we fail instead?
        if self.is_immediate:
            offset = self.value
            if idaapi.is_loaded(offset):
                self._cpu_context.mem_write(offset, value)
            return

        if self.is_register:
            # Convert the value from string to integer...
            if isinstance(value, str):
                value = utils.struct_unpack(value)
            self._cpu_context.registers[self.text] = value
            return

        if self.is_memory_reference:
            # FIXME: Usage of numpy is most likely symptomatic of a bug in an opcode
            #   implementation passing in bad data.
            #   Update this to just is "isinstance" and then fix the buggy opcode.
            # For data written to the frame or memory, this data MUST be a byte string.
            if numpy.issubdtype(type(value), numpy.integer):
                value = utils.struct_pack(value, width=self.width)
            self._cpu_context.mem_write(self.addr, value)
            return

        raise FunctionTracingError(f"Invalid operand type: {self.type}", ip=self.ip)


# This is a "lite" version of the Operand class that only allows access only to a few attributes, is read only,
# and not backed by a CPU context.
OperandLite = collections.namedtuple("OperandLite", "ip idx text value")



"""
This file contains utility functions utilized throughout the function_tracing package.
"""
from __future__ import annotations

import logging
import re
import string
import struct
from typing import TYPE_CHECKING

import ida_funcs
import ida_hexrays
import ida_ida
import ida_idp
import ida_nalt
import ida_typeinf
import idaapi
import idc

from .exceptions import FunctionTracingError
from ..utils import is_x86_64, is_ARM  # TODO: Remove usage where possible.

if TYPE_CHECKING:
    from .operands import Operand


logger = logging.getLogger(__name__)


class SetTypeException(Exception):
    pass


def signed(n, bit_width=None):
    """
    Convert an unsigned integer to a signed integer

    :param uint n: value to convert
    :param int bit_width: bit width of n
        Defaults to architecture addressing size.

    :return int: signed conversion
    """
    if bit_width is None:
        bit_width = get_bits()
    if n >> (bit_width - 1):  # Is the hi-bit set?
        return n - (1 << bit_width)

    return n


def unsigned(n, bit_width=None):
    """
    Convert a signed integer to an unsigned integer

    :param sint n: value to convert
    :param in bit_width: bit width of n
        Defaults to architecture addressing size.

    :return int: unsigned conversion
    """
    if bit_width is None:
        bit_width = get_bits()
    return n & ((1 << bit_width) - 1)


def sign_bit(value, width):
    """Returns the highest bit with given value and byte width."""
    return (value >> ((8 * width) - 1)) & 0x1


def sign_extend(value, orig_size, dest_size):
    """
    Calculates the sign extension for a provided value and a specified destination size.

    :param value: value to be sign extended

    :param orig_size: width of value in bytes

    :param dest_size: size of the destination in bytes

    :return: value, sign extended
    """
    # Calculate the max value for orig and dest
    orig_max = get_mask(orig_size)
    dest_max = get_mask(dest_size)
    # Calculate bit count to shift by
    orig_shift = 8 * orig_size
    dest_shift = 8 * dest_size
    # Create the bit mask
    masknumber = value & orig_max
    msb = masknumber >> (orig_shift - 1)
    # Perform the sign extension
    if msb:
        signextended = ((dest_max << orig_shift) | masknumber) & dest_max
    else:
        signextended = value & dest_max

    return signextended


def align_page_up(x):
    return (x + 0x1000 - 1) & ~(0x1000 - 1)


def get_byte_width(value):
    """
    Calculate the appropriate byte width of the input value (BYTE, WORD, DWORD, QWORD)

    :param value: value to to determine width of

    :return: bytes required to store data
    """
    if value <= 0xFF:
        return 1

    if value <= 0xFFFF:
        return 2

    if value <= 0xFFFFFFFF:
        return 4

    if value <= 0xFFFFFFFFFFFFFFFF:
        return 8


BIG_ENDIAN = ida_ida.inf_is_be()

# Table used by struct_unpack and struct_pack functions
struct_unpack_table = {
    1: "B",
    2: "H",
    4: "L",
    8: "Q",
    16: "QQ",
}


def _get_format_str(width, signed=False):
    if width not in struct_unpack_table:
        raise FunctionTracingError("Invalid width to unpack: {}".format(width))

    format_char = struct_unpack_table[width]
    if signed:
        format_char = format_char.lower()
    return ">" if BIG_ENDIAN else "<" + format_char


def struct_unpack(buffer, signed=False):
    """
    Unpack a buffer given its length and offset using struct.unpack().
    This function will know how to unpack the given buffer by using the lookup table 'struct_unpack_table'
    If the buffer is of unknown length then None is returned. Otherwise the unpacked value is returned.

    :param buffer: data to be unpacked
    :param signed: whether the data is a signed or unsigned value

    :return: unpacked int or None
    """
    n = len(buffer)
    format_str = _get_format_str(n, signed)

    # Unpack
    if n == 16:
        # Struct can only handle up to 64-bit values...so work in 64-bit chunks
        a, b = struct.unpack(format_str, buffer)
        if BIG_ENDIAN:
            return (a << 64) | b
        else:
            return (b << 64) | a
    else:
        return struct.unpack(format_str, buffer)[0]


def struct_pack(value, width=None):
    """
    Pack the supplied value into a string.

    :param value: value to be packed
    :param signed: whether the data is a signed or unsigned value
    :param width: Width of value to pack data into. (choose from 1, 2, 4, 8, or 16 bytes)

    :return: packed data or None
    """
    # FIXME: get_byte_width() is to just keep backwards compatibility. Remove this.
    width = width or get_byte_width(value)
    # Because our & instructions essentially convert our value to a positive integer anyway, we need not worry about
    # whether the input value is signed or not
    # struct.pack("<q", -723873873)
    # '\xaf\x8f\xda\xd4\xff\xff\xff\xff'
    # struct.pack("<Q", -723873873 & 0xFFFFFFFFFFFFFFFF)
    # '\xaf\x8f\xda\xd4\xff\xff\xff\xff'
    format_str = _get_format_str(width, signed=False)

    # Pack
    if width == 16:
        # Struct can only handle up to 64-bit values... so work in 64-bit chunks
        if BIG_ENDIAN:
            return struct.pack(format_str, value >> 64, value & 0xFFFFFFFFFFFFFFFF)
        else:
            return struct.pack(format_str, value & 0xFFFFFFFFFFFFFFFF, value >> 64)
    else:
        # Need to mask off the size of value to width since pack doesn't truncate...
        return struct.pack(format_str, (value & get_mask(width)))


def float_to_int(val, precision=2):
    """
    Given a float value, convert it to its integer hexadecimal equivalent.

    >>> float_to_int(1.0, 8)
    >>> 4607182418800017408

    :param float val: float to convert to int equivalent
    :param int precision: single or double precision (1 for single, 2 for double)
    :return: int
    :raises: ValueError
    """
    if precision == 1:
        return struct.unpack("H", struct.pack("f", val))[0]
    elif precision == 2:
        return struct.unpack("Q", struct.pack("d", val))[0]
    else:
        raise FunctionTracingError("Precision {} is not valid.".format(precision))


def int_to_float(val, precision=2):
    """
    Given an integer value, convert it to its float hexadecimal equivalent.

    >>> int_to_float(4607182418800017408, 8)
    >>> 1.0

    :param int val: integer value to convert to float equivalent
    :param int precision: single or double precision
    :return: int or None
    :raises: ValueError
    """
    if precision == 1:
        return struct.unpack("f", struct.pack("H", val))[0]
    elif precision == 2:
        return struct.unpack("d", struct.pack("Q", val))[0]
    else:
        raise FunctionTracingError("Precision {} is not valid.".format(precision))


def get_mask(size):
    """
    Get bit mask based on byte size.

    :param size: number of bytes to obtain mask for

    :return: mask of width size
    """
    return (1 << (8 * size)) - 1


def get_bits():
    """
    Gets the architecture of the input file.

    :return int: bit width
    """
    info = idaapi.get_inf_structure()
    result = 16
    if info.is_64bit():
        result = 64
    elif info.is_32bit():
        result = 32

    return result


def _decompile(offset: int) -> tuple:
    """
    Attempt to decompile the function at the provided offset using the Hex-Rays plugin. Returns a tuple containing the
    decompiled text (which will be None if an error occurred) and the populated hexrays_failure_t object.
    :param offset: an offset in the function to decompile
    :return: (decompiled text, failure object)
    """
    func = ida_funcs.get_func(offset)
    hf = ida_hexrays.hexrays_failure_t()
    decompiled = ida_hexrays.decompile_func(func, hf, 0)
    return decompiled, hf


def _set_decompiled_func_type(offset: int, decompiled: "ida_hexrays.cfuncptr_t"):
    """
    Given the cfuncptr_t object, obtain the declaration, sanitize it, and set the function's type.
    :param offset: offset of function to set type for
    :param decompiled: a cfuncptr_t object
    """
    # Save type for next time.
    fmt = decompiled.print_dcl()
    fmt = "".join(c for c in fmt if c in string.printable and c not in ("\t", "!"))
    set_type_result = idc.SetType(offset, "{};".format(fmt))
    if not set_type_result:
        logger.warning("Failed to SetType for function at 0x{:X} with decompiler type {!r}".format(offset, fmt))


DECOMPILE_ERRORS = [
    -12,    # call analysis failed
]


def _get_decompiled_function(offset: int) -> "ida_hexrays.cfuncptr_t":
    """
    Attempt to decompile the function containing offset and return the obtained cfuncptr_t object. Additionaly sets the
    type for all decompiled functions.
    :param offset: offset of interest
    :return: a cfuncptr_t object or None
    """
    # This requires Hexrays decompiler, load it and make sure it's available before continuing.
    if not ida_hexrays.init_hexrays_plugin():
        if is_x86_64():
            idc.load_and_run_plugin("hexrays", 0) or idc.load_and_run_plugin("hexx64", 0)
        else:
            idc.load_and_run_plugin("hexarm", 0) or idc.load_and_run_plugin("hexarm64", 0)
    if not ida_hexrays.init_hexrays_plugin():
        raise RuntimeError("Unable to load Hexrays decompiler.")

    decompiled = None
    offsets_to_decompile = [offset] # LIFO list of offsets to try to decompile
    offsets_attempted = set()       # Offsets already attempted to decompile, whether successful or not
    offsets_decompiled = set()      # Set of offsets that successfully decompiled
    # Continue trying to decompile until the list of offsets to decompile is empty or we get an error we can't handle
    # TODO: Determine what errors we can actually handle
    while offsets_to_decompile:
        _offset = offsets_to_decompile.pop()
        # This check means we have probably hit a snag preventing us from decompiling all together, and should prevent
        # an endless loop of continuously trying to decompile
        if _offset in offsets_decompiled:   # already decompiled, but still causing a problem?
            raise RuntimeError("Unable to decompile function 0x{:X}".format(offset))

        offsets_attempted.add(_offset)
        decompiled, hf = _decompile(_offset)
        if decompiled and not hf.code:  # successful decompilation, add to set of decompiled, and remove from attempted
            # set the type
            _set_decompiled_func_type(_offset, decompiled)
            offsets_decompiled.add(_offset)
            offsets_attempted.remove(_offset)
        else:   # unsuccessful...
            if not hf.code in DECOMPILE_ERRORS: # cannot possibly recover
                raise RuntimeError("Unable to decompile function at 0x{:X}".format(offset))

            # Can possibly recover by decompiling called functions
            # add the current offset back to the LIFO list to try decompiling a second time
            offsets_to_decompile.append(_offset)
            # get the address where the analysis failed, pull its operand, and add to the stack
            call_ea = idc.get_operand_value(hf.errea, 0)
            # can't continue if this address was already attempted
            if call_ea in offsets_attempted:
                raise RuntimeError("Unable to decompile function at 0x{:X}".format(offset))

            offsets_to_decompile.append(call_ea)

    return decompiled


def _get_function_tif_with_hex_rays(offset):
    """
    Attempt to get the tinfo_t object of a function using the Hex-Rays decompiler plugin.

    :param offset: Offset of function.
    :raises: RuntimeError on failure.
    :returns: tinfo_t object on success.
    """
    tif = ida_typeinf.tinfo_t()
    decompiled = _get_decompiled_function(offset)
    if not decompiled: # not sure what for shenanigans happened to get None back....
        raise RuntimeError("Expected cfuncptr_t object, received None")

    decompiled.get_func_type(tif)
    return tif


def _get_function_tif_with_guess_type(offset):
    """
    Attempt to get the tinfo_t object of a function using the "guess_type" function.

    :param offset: Offset of function.
    :raises: RuntimeError on failure.
    :returns: tinfo_t object on success.
    """
    tif = ida_typeinf.tinfo_t()

    guessed_type = idc.guess_type(offset)
    if guessed_type is None:
        raise RuntimeError("Failed to guess function type for offset 0x{:X}".format(offset))

    func_name = idc.get_func_name(offset)
    if func_name is None:
        raise RuntimeError("Failed to get function name for offset 0x{:X}".format(offset))

    # Documentation states the type must be ';' terminated, also the function name must be inserted
    guessed_type = re.sub(r"\(", " {}(".format(func_name), "{};".format(guessed_type))
    set_type_result = idc.SetType(offset, guessed_type)
    if not set_type_result:
        logger.warning(
            "Failed to SetType for function at 0x{:X} with guessed type {!r}".format(offset, guessed_type)
        )
    # Try one more time to get the tinfo_t object
    if not ida_nalt.get_tinfo(tif, offset):
        raise RuntimeError("Failed to obtain tinfo_t object for offset 0x{:X}".format(offset))

    return tif


# Cache of function types we have computed.
_func_types = set()


def get_function_data(offset, operand: Operand = None):
    """
    Obtain a idaapi.func_type_data_t object for the function with the provided start EA.

    :param int offset: start EA of function
    :param operand: operand containing function address in it's value.
        This can be provided when function is dynamically generated at runtime. (e.g. call eax)

    :return: ida_typeinf.func_type_data_t object, ida_typeinf.tinfo_t object

    :raise RuntimeError: if func_type_data_t object cannot be obtained
    """
    global _func_types

    tif = None

    try:
        func_type = idc.get_type(offset)
    except TypeError:
        raise RuntimeError("Not a valid offset: {!r}".format(offset))

    # First see if it's a type we already set before.
    if func_type and offset in _func_types:
        tif = ida_typeinf.tinfo_t()
        ida_nalt.get_tinfo(tif, offset)

    else:
        # Otherwise, try to use the Hexrays decompiler to determine function signature.
        # (It's better than IDA's guess_type)
        try:
            tif = _get_function_tif_with_hex_rays(offset)

        # If we fail, resort to using guess_type+
        except RuntimeError:
            if func_type:
                # If IDA's disassembler set it already, go with that.
                tif = ida_typeinf.tinfo_t()
                ida_nalt.get_tinfo(tif, offset)
            else:
                try:
                    tif = _get_function_tif_with_guess_type(offset)
                except RuntimeError:
                    # Don't allow to fail if we could pull from operand.
                    pass

    if tif:
        funcdata = ida_typeinf.func_type_data_t()

        # In IDA 7.6, imported functions are now function pointers.
        # To handle this, check if we need to pull out a pointed object first
        if tif.is_funcptr():
            tif = tif.get_pointed_object()

        success = tif.get_func_details(funcdata)
        if success:
            # record that we have processed this function before. (and that we can grab it from the offset)
            _func_types.add(offset)
            return funcdata, tif

    # If we have still failed, we have one more trick under our sleeve.
    # Try to pull the type information from the operand of the call instruction.
    # This could be set if the function has been dynamically created.
    if operand:
        tif = operand._tif
        funcdata = ida_typeinf.func_type_data_t()
        success = tif.get_func_details(funcdata)
        if success:
            return funcdata, tif

    raise RuntimeError("failed to obtain func_type_data_t object for offset 0x{:X}".format(offset))


def get_function_name(func_ea):
    """Retrieves the function name from the given address."""
    # (Using get_name() over get_func_name() so it also works for imported functions)
    func_name = idc.get_name(func_ea)
    # Demangle name if necessary:
    demangled_name = idc.demangle_name(func_name, idc.get_inf_attr(idc.INF_SHORT_DN))
    if demangled_name:
        # Strip off the extra junk: 'operator new(uint)' -> 'new'
        match = re.search("([^ :]*)\(", demangled_name)
        if not match:
            logger.debug("Unable to demangle function name: {}".format(demangled_name))
        else:
            logger.debug("Demangled function name {} -> {}".format(demangled_name, match.group(1)))
            demangled_name = match.group(1)
        func_name = demangled_name
    return func_name


def sanitize_func_name(func_name):
    """Sanitizes the IDA function names to it's core name."""
    # remove the extra "_" IDA likes to add to the function name.
    if func_name.startswith("_"):
        func_name = func_name[1:]

    # Remove the numbered suffix IDA likes to add to duplicate function names.
    func_name = re.sub("_[0-9]+$", "", func_name)

    return func_name


def is_func_ptr(offset: int) -> bool:
    """Returns true if the given offset is a function pointer."""
    # As a first check, simply see if the offset is the start of a function.
    func = ida_funcs.get_func(offset)
    if func and func.start_ea == offset:
        return True

    # Sometimes we will get a really strange issue where the IDA disassember has set a type for an
    # address that should not have been set during our course of emulation.
    # Therefore, before attempting to use get_function_data() to test if it's a function pointer,
    # first see if guess_type() will return None while is_loaded() is true.
    # If it doesn't, we know that it shouldn't be a function pointer.
    # (plus it saves on time)
    # TODO: Determine if we could have false negatives.
    #   - this caused a false negatives, so I added the check if offset is the start of a function.
    try:
        if idc.is_loaded(offset) and not idc.guess_type(offset):
            return False
    except TypeError:
        return False
    try:
        get_function_data(offset)
        return True
    except RuntimeError:
        return False
    except Exception as e:
        # If we get any other type of exception raise a more friendly error message.
        raise FunctionTracingError("Failed to retrieve function data from {!r}: {}".format(offset, e))


def convert_reg(reg_name, width):
    """Convert given register name to the register name for the provided width (eg: conver_reg(eax, 8) -> rax)"""
    reg_idx = ida_idp.str2reg(reg_name)
    if reg_idx > 15:  # R15 is index 15.  8-bit registers have indexes above 15 but are 0 indexed, so sub 16
        reg_idx -= 16

    return ida_idp.get_reg_name(reg_idx, width)


def reg2str(register, width=None):
    """Convert given register index to the register name with the provided width (eg: reg2str(0, 8) -> rax)"""
    if not width:
        width = 8 if idc.__EA64__ else 4
    return ida_idp.get_reg_name(register, width)



import logging
import itertools

import idc
import idaapi
import idautils


from kordesii import logutil


logger = logging.getLogger(__name__)


def try_make_function(
    function_start, function_end=idc.BADADDR, target_location=None, require_term=True, end_mnem_bytes=None
):
    """
    Description:
        Given a function location, attempt to create a function.
        If function creation fails, delete any partially created functions.
        If function creation succeeds, ensure all of the function's bytes are analyzed as code.

    Input:
        function_start - The start_ea of the function to create
        function_end - The end_ea of the function to create. IDA will calculate if not provided.
        target_location - If provided, fail function creation if it does not include this EA
        require_term - If provided, fail function creation if the last instruction is not a ret or jmp
        end_mnem_bytes - If provided, fail function creation if the last instruction is not the provided bytes
                         Instructions are entered as space separated bytes (i.e. '55' for 'push ebp')

    Output:
        Returns a tuple (function_start, function_end) for the created function if successful, None otherwise
    """
    if function_start <= function_end:
        if idc.add_func(function_start, function_end):
            logger.debug("Created a function 0x%X - 0x%X." % (function_start, function_end))
            if require_term:
                last_mnem_ea = idc.get_item_head(idaapi.get_func(function_start).end_ea - 1)
                last_mnem = idc.print_insn_mnem(last_mnem_ea)
                if (end_mnem_bytes is None and "ret" not in last_mnem and "jmp" not in last_mnem) or (
                    end_mnem_bytes
                    and idc.get_bytes(last_mnem_ea, idc.get_item_size(last_mnem_ea)).encode("hex").upper()
                    != end_mnem_bytes.upper()
                ):
                    idc.del_func(function_start)
                    logger.debug(
                        "Deleted function at 0x%X - the function didn't end with the correct mnem/bytes."
                        % function_start
                    )
                    return
            if target_location is not None:
                if function_start <= target_location < idaapi.get_func(function_start).end_ea:
                    idc.plan_and_wait(function_start, idaapi.get_func(function_start).end_ea)
                    return function_start, function_end
                else:
                    idc.del_func(function_start)
                    logger.debug(
                        "Deleted function at 0x%X - the function didn't contain the target location." % function_start
                    )
                    return
        else:
            logger.debug(
                "Tried to create a function 0x%X - 0x%X, but IDA wouldn't do it." % (function_start, function_end)
            )
    else:
        logger.debug("The end address was not greater than the start address!")


def find_binary_instruction_start(
    search_start_location,
    search_direction,
    target,
    min_location=idc.get_inf_attr(idc.INF_MIN_EA),
    max_location=idc.get_inf_attr(idc.INF_MAX_EA),
):
    """
    Description:
        Given a starting location, target, and direction, find an instruction starting with the target bytes.

    Input:
        search_start_location - The EA to start searching at
        search_direction - either idc.SEARCH_UP or idc.SEARCH_DOWN
        target - The target as space separated bytes (i.e. '55' for 'push ebp')
        min_location - The minimum EA to accept results for (default: idc.get_inf_attr(idc.INF_MIN_EA))
        max_location - The maximum EA to accept results for (default: idc.get_inf_attr(idc.INF_MAX_EA))

    Output:
        Returns the first matching location if found, otherwise idc.BADADDR
    """
    target = target.upper()
    while search_start_location < max_location:
        ea = idc.find_binary(search_start_location, search_direction, target)
        if (
            min_location <= ea < max_location
            and ea == idc.get_item_head(ea)
            and idc.get_bytes(ea, idc.get_item_size(ea)).hex().upper().startswith(target.replace(" ", ""))
        ):
            return ea
        else:
            search_start_location = ea + (1 if search_direction == idc.SEARCH_DOWN else -1)
    return idc.BADADDR


def calc_most_common_start_bytes():
    """
    Description:
        Iterate over all non-lib (and non-thunk) functions and record their first instruction.
        Return the bytes for whichever instruction appears most.

    Output:
        A space separated string of bytes of the most common first instruction.
    """
    counts = {}
    for func_ea in idautils.Functions():
        if not idc.get_func_attr(func_ea, idc.FUNCATTR_FLAGS) & (idc.FUNC_LIB | idc.FUNC_THUNK):
            start_bytes = idc.get_bytes(func_ea, idc.get_item_size(func_ea))
            if start_bytes in counts:
                counts[start_bytes] += 1
            else:
                counts[start_bytes] = 1
    return " ".join("%02X" % ord(c) for c in sorted(list(counts.items()), key=lambda tup: tup[1], reverse=True)[0][0])


def sanity_checks(location):
    """
    Description:
        Do some basic checks to see if a function can be created containing the provided EA.

    Input:
        location - The EA to evaluate

    Output:
        True if a function can be created containing the provided EA
        False if a the provided EA was a nop or Align
        None if there is already a function containing the provided EA
    """
    if idaapi.get_func(location):
        logger.debug("There's already a function here! (0x%X)" % location)
        return None
    elif (
        idc.is_align(idc.get_full_flags(location))
        or idc.print_insn_mnem(location) == "nop"
        or (idaapi.is_data(idc.get_full_flags(location)) and idc.get_wide_byte(location) == 0x90)
    ):
        # Yes, the nop bit may be incorrect, but it's gonna be a very special case that needs a function with nops
        logger.warning("Can't make a function including aligns and/or nops!")
        return False
    else:
        return True


def trim_func(ea, GetHead):
    """
    Description:
        Steps until it hits something not a nop or not starts with 90 (nop opcode) nor an align or not byte 0xCC (Align 'opcode').

    Input:
        ea - The location to adjust for nops and Aligns. EA must be a head.
        GetHead - either prev_head or next_head

    Output:
        The corrected EA.
    """
    while (
        idc.print_insn_mnem(ea) == "nop"
        or (idaapi.is_data(idc.get_full_flags(ea)) and idc.get_wide_byte(ea) == 0x90)
        or idc.is_align(idc.get_full_flags(ea))
        or (not idc.is_code(idc.get_full_flags(ea)) and idc.get_wide_byte(ea) == 0xCC)
    ):
        ea = GetHead(ea)
    return ea


def find_function_starts_near(location, start_mnem_bytes=None):
    """
    Description:
        Identifies the nearest possible function starts since the most recent function or Align.

    Input:
        location - The EA to search before
        start_mnem_bytes - Try to start functions on a particular instruction
                           Instructions are entered as space separated bytes (i.e. '55' for 'push ebp')
                           The specified pattern will be used first, then the defaults will be used
                           If no pattern is specified, the defaults will be used, which prefers 'push ebp'

    Output:
        starts - A list of function end EAs sorted: start_mnem_bytes, push ebp, (push esp, push esi, push edi)
    """
    # foreach target bytes:
    #  step instructions up
    #  if instruction matches the target bytes, add to output list
    #   then move on to the next target bytes
    #  if we hit a function or an align, quit
    # return starts in the order
    #  start_nmem_bytes
    #  push ebp
    #  others, sorted descending

    min_location = None
    ea = location
    while min_location is None:
        ea = idc.prev_head(ea)
        if idaapi.get_func(ea) or idc.is_align(idc.get_full_flags(ea)):
            min_location = ea
        elif ea == idc.BADADDR:
            min_location = idaapi.getseg(location).start_ea
    min_location = max(min_location, idaapi.getseg(location).start_ea)

    targets = ["55", "54", "56", "57"]
    if start_mnem_bytes:
        targets.insert(0, start_mnem_bytes)

    starts = {}
    for target in targets:
        ea = find_binary_instruction_start(location - 1, idc.SEARCH_UP, target, min_location)
        if ea != idc.BADADDR:
            starts[target] = ea

    return [
        start
        for start in (
            [starts.get(start_mnem_bytes, None), starts.get("55", None)]
            + sorted([starts.get(target, -1) for target in targets[-3:]], reverse=True)
        )
        if start
    ]


def find_function_starts(location, start_mnem_bytes=None):
    """
    Description:
        Identifies all possible function starts since the most recent function or Align.

    Input:
        location - The EA to search before
        start_mnem_bytes - Try to start functions on a particular instruction
                           Instructions are entered as space separated bytes (i.e. '55' for 'push ebp')
                           The specified pattern will be used first, then the defaults will be used
                           If no pattern is specified, the defaults will be used, which prefers 'push ebp'

    Output:
        starts - A list of function end EAs sorted: start_mnem_bytes, push ebp, (push esp, push esi, push edi)
    """
    # foreach target bytes:
    #  step instructions up
    #  if instruction matches the target bytes, add to the corresponding output list
    #  if we hit a function or an align, quit
    # return starts in the order:
    #  start_nmem_bytes
    #  push ebp
    #  others, sorted ascending

    min_location = None
    ea = location
    while min_location is None:
        ea = idc.prev_head(ea)
        if idaapi.get_func(ea) or idc.is_align(idc.get_full_flags(ea)):
            min_location = ea
        elif ea == idc.BADADDR:
            min_location = idaapi.getseg(location).start_ea
    min_location = max(min_location, idaapi.getseg(location).start_ea)

    targets = ["55", "54", "56", "57"]
    if start_mnem_bytes:
        targets.insert(0, start_mnem_bytes)

    starts = {}
    for target in targets:
        function_starts = []
        ea = find_binary_instruction_start(location - 1, idc.SEARCH_UP, target, min_location)
        while ea != idc.BADADDR:
            if ea < min_location:
                break
            else:
                function_starts.append(ea)
            ea = find_binary_instruction_start(ea - 1, idc.SEARCH_UP, target, min_location)
        starts[target] = function_starts

    return (
        (starts[start_mnem_bytes] if start_mnem_bytes else [])
        + starts["55"]
        + sorted(itertools.chain.from_iterable(starts[target] for target in targets[-3:]))
    )


def find_function_ends_near(location, end_mnem_bytes=None):
    """
    Description:
        Identifies the nearest possible function ends before the next function or Align for each end mnem.

    Input:
        location - The EA to search after
        end_mnem_bytes - Try to end functions on a particular instruction
                         Instructions are entered as space separated bytes (i.e. 'C2' for 'retn')
                         The specified pattern will be used first, then the defaults will be used
                         If no pattern is specified, the defaults will be used, which prefers 'retn'

    Output:
        ends - A list of function end EAs sorted: end_mnem_bytes, retn, jmp
    """
    # foreach target bytes:
    #  step instructions down
    #  if instruction matches the target bytes, add to output list
    #   then move on to the next target bytes
    #  if we hit a function or an align, quit
    # return ends in the order
    #  end_nmem_bytes
    #  retn
    #  jmp
    #  others, sorted ascending

    max_location = None
    ea = location
    while max_location is None:
        ea = idc.next_head(ea)
        if idaapi.get_func(ea) or idc.is_align(idc.get_full_flags(ea)):
            max_location = ea
        elif ea == idc.BADADDR:
            max_location = idaapi.getseg(location).end_ea
    max_location = min(max_location, idaapi.getseg(location).end_ea)

    targets = ["C2", "C3", "E9", "EA", "EB"]
    if end_mnem_bytes:
        targets.insert(0, end_mnem_bytes)

    ends = {}
    for target in targets:
        ea = find_binary_instruction_start(location, idc.SEARCH_DOWN, target, max_location=max_location)
        if ea <= max_location:
            ends[target] = ea

    return [
        end + idc.get_item_size(end)
        for end in (
            ([ends.get(end_mnem_bytes, None), ends.get("C2", None), ends.get("C3", None)])
            + sorted(ends.get(target, -1) for target in targets[-3:])
        )
        if end
    ]


def find_function_ends(location, end_mnem_bytes=None):
    """
    Description:
        Identifies all possible function ends before the next function or Align.

    Input:
        location - The EA to search after
        end_mnem_bytes - Try to end functions on a particular instruction
                         Instructions are entered as space separated bytes (i.e. 'C2' for 'retn')
                         The specified pattern will be used first, then the defaults will be used
                         If no pattern is specified, the defaults will be used, which prefers 'retn'

    Output:
        ends - A list of function end EAs sorted: end_mnem_bytes, retn, jmp
    """
    # foreach target bytes:
    #  step instructions down
    #  if instruction matches the target bytes, add to the corresponding output list
    #  if we hit a function or an align, quit
    # return ends in the order:
    #  end_nmem_bytes
    #  retn
    #  jmp
    #  others, sorted ascending

    max_location = None
    ea = location
    while max_location is None:
        ea = idc.next_head(ea)
        if idaapi.get_func(ea) or idc.is_align(idc.get_full_flags(ea)):
            max_location = ea
        elif ea == idc.BADADDR:
            max_location = idaapi.getseg(location).end_ea
    max_location = min(max_location, idaapi.getseg(location).end_ea)

    targets = ["C3", "C2", "E9", "EA", "EB"]
    if end_mnem_bytes:
        targets.insert(0, end_mnem_bytes)

    ends = {}
    for target in targets:
        function_ends = []
        ea = find_binary_instruction_start(location, idc.SEARCH_DOWN, target, max_location=max_location)
        while ea != idc.BADADDR:
            if ea > max_location:
                break
            else:
                function_ends.append(ea)
            ea = find_binary_instruction_start(ea + 11, idc.SEARCH_DOWN, target, max_location=max_location)
        ends[target] = function_ends

    return [
        end + idc.get_item_size(end)
        for end in (
            (ends[end_mnem_bytes] if end_mnem_bytes else [])
            + sorted(ends["C3"] + ends["C2"])
            + sorted(itertools.chain.from_iterable(ends[target] for target in targets[-3:]))
        )
    ]


def create_function_here(location, require_term=True, end_mnem_bytes=None):
    """
    Description:
        Attempt to make a function starting at the provided EA. First, try to have IDA find the end.
        If that fails, try to find the end ourselves.

    Input:
        location - The EA at which IDA should attempt to make a function.
        require_term - When True, requires the last instruction in all defined functions to be retn or jmp
        end_mnem_bytes - Try to end functions on a particular instruction
                         Instructions are entered as space separated bytes (i.e. 'C2' for 'retn')
                         The specified pattern will be used first, then the defaults will be used
                         If no pattern is specified, the defaults will be used, which prefers 'retn'

    Output:
        True if it made a function or a function was already present, False otherwise.
    """
    sanity = sanity_checks(location)
    if sanity is None:  # There was already a function
        return True
    elif sanity is False:  # There was something preventing function creation
        return False

    function_ends = find_function_ends(location, end_mnem_bytes)
    function_ends.insert(0, idc.BADADDR)
    for end in function_ends:
        if try_make_function(location, end, location, require_term, end_mnem_bytes):
            return True
    return False


def create_function_precise(location, require_term=True, start_mnem_bytes=None, end_mnem_bytes=None):
    """
    Description:
        Attempt to make a function containing <location> and only that function.
        First tries to let IDA find the end of the calculated start EA.
        If that fails, try to calculate the end ourselves.

    Input:
        location - An address that should be within a function
        require_term - When True, requires the last instruction in all defined functions to be retn or jmp
        start_mnem_bytes - Try to start functions on a particular instruction
                           Instructions are entered as space separated bytes (i.e. '55' for 'push ebp')
                           The specified pattern will be used first, then the defaults will be used
                           If no pattern is specified, the defaults will be used, which prefers 'push ebp'
        end_mnem_bytes - Try to end functions on a particular instruction
                         Instructions are entered as space separated bytes (i.e. 'C2' for 'retn')
                         The specified pattern will be used first, then the defaults will be used
                         If no pattern is specified, the defaults will be used, which prefers 'retn'

    Output:
        True if it made a function or a function was already present, False otherwise.
    """
    sanity = sanity_checks(location)
    if sanity is None:  # There was already a function
        return True
    elif sanity is False:  # There was something preventing function creation
        return False

    logger.debug("Trying to make function for 0x%X" % location)
    function_starts = find_function_starts_near(location, start_mnem_bytes)
    if not function_starts:
        return False  # If we don't have any start points, we're up a creek

    # Don't populate function_ends at this point to avoid the tracing we aren't sure we need yet
    # This will cause two repeats in the last section if we get that far, but that's an acceptable trade-off

    # Try to make a function at the most likely start point letting IDA calculate the end
    if try_make_function(
        function_starts[0], target_location=location, require_term=require_term, end_mnem_bytes=end_mnem_bytes
    ):
        return True
    else:  # If that fails, try to make a function at that point with the most likely end
        function_ends = find_function_ends_near(location, end_mnem_bytes)
        # Only try the first end here. This guarantees that one of the lower tier starts won't work with idc.BADADDR before we try this end
        if function_ends and try_make_function(
            function_starts[0], function_ends[0], location, require_term, end_mnem_bytes
        ):
            return True

    # Always let IDA have the first shot at finding the end for each start
    function_ends.insert(0, idc.BADADDR)
    # For each end, try each start, that way each start gets a shot at the most likely end before we try the next most likely one
    for function_end in function_ends:
        # For each start, try to make a function with the current end
        for function_start in function_starts:
            if try_make_function(function_start, function_end, location, require_term, end_mnem_bytes):
                return True
    return False


def ida_make_functions(location, require_term=True):
    """
    Description:
        Attempts to create functions based on the assumption that there should be continuous contiguous
        functions defined since the previous function or align. Stops creating functions once a function
        containing <location> is created or the next created function would be past <location>.
        Only identifies potential start EAs and lets IDA find the ends.

    Input:
        location - The EA at which IDA should attempt to make a function.
        require_term - When True, requires the last instruction in all defined functions to be retn or jmp

    Output:
        True if it made a function or a function was already present, False otherwise.
    """
    sanity = sanity_checks(location)
    if sanity is None:  # There was already a function
        return True
    elif sanity is False:  # There was something preventing function creation
        return False

    target_location = location
    function_start = location
    ea = location
    while not (idaapi.get_func(ea) or idc.is_align(idc.get_full_flags(ea))):
        function_start = ea
        ea = idc.prev_head(ea)
    function_start = trim_func(function_start, idc.next_head)

    if try_make_function(function_start, require_term=require_term):
        if not idaapi.get_func(target_location):
            return ida_make_functions(target_location, require_term)
        else:
            return True
    else:
        return False


def create_functions(location, require_term=True, start_mnem_bytes=None, end_mnem_bytes=None):
    """
    Description:
        Attempts to create functions based on the assumption that there should be continuous contiguous
        functions defined since the previous function or align. Stops creating functions once a function
        containing <location> is created or the next created function would be past <location>.
        Finds both start and end EAs, not using on IDA's algorithms.

    Input:
        location - An address that should be within a function
        require_term - When True, requires the last instruction in all defined functions to be retn or jmp
        start_mnem_bytes - Try to start functions on a particular instruction
                           Instructions are entered as space separated bytes (i.e. '55' for 'push ebp')
                           The specified pattern will be used first, then the defaults will be used
                           If no pattern is specified, the defaults will be used, which prefers 'push ebp'
        end_mnem_bytes - Try to end functions on a particular instruction
                         Instructions are entered as space separated bytes (i.e. 'C2' for 'retn')
                         The specified pattern will be used first, then the defaults will be used
                         If no pattern is specified, the defaults will be used, which prefers 'retn'

    Output:
        True if it made a function or a function was already present, False otherwise.
    """
    sanity = sanity_checks(location)
    if sanity is None:  # There was already a function
        return True
    elif sanity is False:  # There was something preventing function creation
        return False

    # Attempt to find the function ourselves.
    function_starts = find_function_starts(location, start_mnem_bytes)
    function_ends = find_function_ends(location, end_mnem_bytes)

    found_func = None
    for function_start, function_end in itertools.product(function_starts, function_ends):
        if function_start < function_end:
            if (
                try_make_function(function_start, function_end, require_term=require_term)
                and function_start <= location < idaapi.get_func(function_start).end_ea
            ):
                found_func = (function_start, function_end)
                break  # Don't return here in case we have to split it yet.

    if found_func:
        return True
    else:
        logger.warning("Failed to find function based on location 0x%X." % location)
        return False



"""
IDA logging utility.
"""

import copy
import errno
import inspect
import logging
import logging.config
import logging.handlers
import os
import struct
import sys
import threading
import traceback
import warnings
from collections import deque

import kordesii
import kordesii.config as kordesii_config
import yaml
from six.moves import socketserver, cPickle as pickle


class LevelCharFilter(logging.Filter):
    """Logging filter used to add a 'level_char' format variable."""

    def filter(self, record):
        if record.levelno >= logging.ERROR:
            record.level_char = "!"
        elif record.levelno >= logging.WARN:
            record.level_char = "-"
        elif record.levelno >= logging.INFO:
            record.level_char = "+"
        elif record.levelno >= logging.DEBUG:
            record.level_char = "*"
        else:
            record.level_char = " "
        return True


class MPRotatingFileHandler(logging.handlers.RotatingFileHandler):
    """
    Handle the uncommon case of the log attempting to roll over when
    another process has the log open. This only happens on Windows, and
    the log ends up being a handful of KBs greater than 1024. Entries
    are still written, and the rollover happens if/when the MainProcess is
    the only process with the log file open.
    """

    def __init__(self, filename, **kwargs):
        # Expand and variables and home directories and make path if it doesn't exist.
        filename = os.path.expandvars(os.path.expanduser(filename))
        directory = os.path.dirname(filename)
        if not os.path.exists(directory):
            os.makedirs(directory)
        super(MPRotatingFileHandler, self).__init__(filename, **kwargs)

    def doRollover(self):
        """
        Attempt to roll over to the next log file. If the current file
        is locked (Windows issue), keep writing to the original file until
        it is unlocked.

        :return:
        """
        try:
            super(MPRotatingFileHandler, self).doRollover()
        except OSError as e:
            if not (sys.platform == "win32" and e.errno == errno.EACCES):
                raise


class ListHandler(logging.Handler):
    """
    Log to a list, with an optional maximum number of records to store.

    Full records are available with the `records` property, and messages (i.e.
    the text of the log entry) at available with the `messages` property.
    """

    def __init__(self, entries=None):
        """
        Behaves essentially identical to any other handler.

        The only option is max_entries, to specify the max number of log
        entries kept. By default, no limit.

        :param int entries: Maximum number of records to store.
        """
        super(ListHandler, self).__init__()

        self._deque = deque(maxlen=entries)

    def __copy__(self):
        new_handler = ListHandler()
        # Actually copy the deque, otherwise we'll get double entries
        new_handler._deque = copy.copy(self._deque)
        return new_handler

    def emit(self, record):
        msg = self.format(record)
        record.formatted_msg = msg
        self._deque.append(record)

    def clear(self):
        return self._deque.clear()

    @property
    def records(self):
        """
        List of the last `max_entries` records logged.
        """
        return list(self._deque)

    @property
    def messages(self):
        """
        List of the last `max_entries` formatted messages logged.
        """
        return [record.formatted_msg for record in self._deque]


class LogRecordStreamHandler(socketserver.StreamRequestHandler):
    """Handler for a streaming logging request.

    This basically logs the record using whatever logging policy is
    configured locally.
    """

    def handle(self):
        """
        Handle multiple requests - each expected to be a 4-byte length,
        followed by the LogRecord in pickle format. Logs the record
        according to whatever policy is configured locally.
        """
        while True:
            chunk = self.connection.recv(4)
            if len(chunk) < 4:
                break
            slen = struct.unpack(">L", chunk)[0]
            chunk = self.connection.recv(slen)
            while len(chunk) < slen:
                chunk = chunk + self.connection.recv(slen - len(chunk))
            obj = self.unpickle(chunk)
            record = logging.makeLogRecord(obj)
            self.handle_log_record(record)

    def unpickle(self, data):
        return pickle.loads(data)

    def handle_log_record(self, record):
        # if a name is specified, we use the named logger rather than the one
        # implied by the record.
        if self.server.logname is not None:
            name = self.server.logname
        else:
            name = record.name
        logger = logging.getLogger(name)
        # TODO: Make sure we still filter on other end to save network bandwidth!
        if logger.isEnabledFor(record.levelno):
            logger.handle(record)


class LogRecordSocketReceiver(socketserver.ThreadingTCPServer):
    """
    Simple TCP socket-based logging receiver.
    """

    allow_reuse_address = 1

    def __init__(self):
        # Since this is all local, we are using port 0 to let the system pick a random open port for us.
        socketserver.ThreadingTCPServer.__init__(self, ("localhost", 0), LogRecordStreamHandler)
        self.abort = 0
        self.timeout = 1
        self.logname = None

    def serve_until_stopped(self):
        import select

        abort = 0
        while not abort:
            rd, wr, ex = select.select([self.socket.fileno()], [], [], self.timeout)
            if rd:
                self.handle_request()
            abort = self.abort


_started_listener = False
# Port being used to listen to logs.
listen_port = None


def start_listener():
    """Start the listener thread for socket-based logging."""
    global _started_listener
    global listen_port

    if _started_listener:
        return

    tcp_server = LogRecordSocketReceiver()
    _, listen_port = tcp_server.server_address

    listener_thread = threading.Thread(target=tcp_server.serve_until_stopped)
    listener_thread.daemon = True
    listener_thread.start()
    # Make sure we only start this once.
    _started_listener = True


_setup_logging = False


def setup_logging(level=None):
    """
    Sets up logging using default log config file or log config file set by 'KORDESII_LOG_CFG'

    :param level: Log level to set.
        If not provided, level will be based on what is currently set in the root logger.
    """
    if kordesii.in_ida:
        if kordesii.called_from_framework:
            # If we are in IDA and part of a framework call, setup up socket handler that will
            # get received by the framework.
            # (log level and port will passed in as the 1st and 2nd command line argument)
            import idc

            log_level = int(idc.ARGV[1])
            log_port = int(idc.ARGV[2])

            logging.root.setLevel(log_level)
            socket_handler = logging.handlers.SocketHandler("localhost", log_port)
            logging.root.addHandler(socket_handler)
        else:
            # If running decoder from IDA interface, send simple logs to output window only.
            global _setup_logging
            if not _setup_logging:
                stream_handler = logging.StreamHandler()  # IDA redirects sys.stderr to output window.
                logging.root.addHandler(stream_handler)
                stream_handler.addFilter(LevelCharFilter())
                stream_handler.setFormatter(logging.Formatter("[%(level_char)s] %(module)-15s : %(message)s"))

                if level:
                    logging.root.setLevel(level)
                # Use INFO level if log level wasn't set by user.
                elif logging.root.getEffectiveLevel() == logging.NOTSET:
                    logging.root.setLevel(logging.INFO)

            # Make sure we only setup once to avoid duplicate log messages.
            _setup_logging = True
    else:
        # Allow setting log configuration using 'KORDESII_LOG_CFG' environment variable.
        log_config = os.getenv("KORDESII_LOG_CFG", kordesii_config.LOG_CONFIG_PATH)
        try:
            with open(log_config, "rt") as f:
                config = yaml.safe_load(f.read())
            logging.config.dictConfig(config)
        except IOError as e:
            warnings.warn("Unable to set log config file: {} with error: {}".format(log_config, e))
            logging.basicConfig(level=level or logging.INFO)

        # Receive decoder logs passed though the socket.
        start_listener()


def get_logger(call_level=1):
    """
    Helper function for getting the correctly named logger for decoders.
    This function is necessary because all decoder modules's __name__ attribute
    will be "__main__" when run in IDA.

    :param int call_level: Call level to pull module name from.
        (defaults to 1 which is the caller of this function)

    Add the following to the top of your decoder:

        import kordesii

        logger = kordesii.get_logger()

    :return: logging.Logger object
    """
    try:
        frame, file_path, _, _, _, _ = inspect.stack(0)[call_level]
        caller_name = frame.f_locals.get("__name__", "__main__")
    except TypeError:
        f = sys._getframe().f_back
        stack = traceback.StackSummary.extract(traceback.walk_stack(f), limit=call_level + 1, capture_locals=True)
        stack.reverse()
        frame = stack[call_level]
        caller_name = frame.locals.get("__name__", "__main__")
        file_path = frame.filename
    if caller_name == "__main__":
        # Manually construct module path based on file locations.
        module_path = os.path.relpath(file_path, os.path.dirname(kordesii.__file__))
        module_path = ["kordesii"] + list(filter(None, list(os.path.split(module_path))))
        module_path[-1] = os.path.splitext(module_path[-1])[0]  # remove extension
        caller_name = ".".join(module_path)

    if os.sep in caller_name:
        caller_name = "kordesii." + caller_name.rpartition(os.sep)[2]

    return logging.getLogger(caller_name)



"""
This is a partial port of IDA's intel.hpp from the Hex-Rays SDK.
(Hex-Rays does not officially expose this module in IDAPython.)
"""
from enum import IntEnum, auto

import ida_allins
import ida_ua
import idc
import idaapi


# region - Copied from kordesii.utils.function_tracing.utils

def signed(n, bit_width):
    """
    Convert an unsigned integer to a signed integer

    :param uint n: value to convert
    :param int bit_width: byte width of n

    :return int: signed conversion
    """
    if n >> (bit_width - 1):  # Is the hi-bit set?
        return n - (1 << bit_width)

    return n


def get_bits():
    """
    Gets the architecture of the input file.

    :return int: bit width
    """
    info = idaapi.get_inf_structure()
    result = 16
    if info.is_64bit():
        result = 64
    elif info.is_32bit():
        result = 32

    return result

# endregion


# Intel 80x86 insn_t.auxpref bits
aux_lock = 0x00000001
aux_rep = 0x00000002
aux_repne = 0x00000004
aux_use32 = 0x00000008         # segment type is 32-bits
aux_use64 = 0x00000010         # segment type is 64-bits
aux_large = 0x00000020         # offset field is 32-bit (16-bit is not enough)
aux_short = 0x00000040         # short (byte) displacement used
aux_sgpref = 0x00000080        # a segment prefix byte is not used
aux_oppref = 0x00000100        # operand size prefix byte is not used
aux_adpref = 0x00000200        # address size prefix byte is not used
aux_basess = 0x00000400        # SS based instruction
aux_natop = 0x00000800         # operand size is not overridden by prefix
aux_natad = 0x00001000         # addressing mode is not overridden by prefix
aux_fpemu = 0x00002000         # FP emulator instruction
aux_vexpr = 0x00004000         # VEX-encoded instruction
aux_bnd = 0x00008000           # MPX-encoded instruction
aux_evex = 0x00010000          # EVEX-encoded instruction
aux_xop = 0x00020000           # XOP-encoded instruction
aux_xacquire = 0x00040000      # HLE prefix hints
aux_xrelease = 0x00080000      # HLE prefix hints


# bits in insn_t.evex_flags:
EVEX_R = 0x01           # High-16 register specifier modifier
EVEX_L = 0x02           # Vector length/RC
EVEX_z = 0x04           # Zeroing/Merging
EVEX_b = 0x08           # Broadcast/RC/SAE Context
EVEX_V = 0x10           # High-16 NDS/VIDX register specifier

# bits in insn_t.rex:
REX_W = 8               # 64-bit operand size
REX_R = 4               # modrm reg field extension
REX_X = 2               # sib index field extension
REX_B = 1               # modrm r/m, sib base, or opcode reg fields extension
VEX_L = 0x80            # 256-bit operation (YMM register)

regnum_t = int


class RegNo(IntEnum):
    R_none = -1
    R_ax = 0
    R_cx = auto() # 1
    R_dx = auto() # 2
    R_bx = auto() # 3
    R_sp = auto() # 4
    R_bp = auto() # 5
    R_si = auto() # 6
    R_di = auto() # 7
    R_r8 = auto() # 8
    R_r9 = auto() # 9
    R_r10 = auto() # 10
    R_r11 = auto() # 11
    R_r12 = auto() # 12
    R_r13 = auto() # 13
    R_r14 = auto() # 14
    R_r15 = auto() # 15

    R_al = auto()
    R_cl = auto()
    R_dl = auto()
    R_bl = auto()
    R_ah = auto()
    R_ch = auto()
    R_dh = auto()
    R_bh = auto()

    R_spl = auto()
    R_bpl = auto()
    R_sil = auto()
    R_dil = auto()

    R_ip = auto()

    R_es = auto() # 0
    R_cs = auto() # 1
    R_ss = auto() # 2
    R_ds = auto() # 3
    R_fs = auto()
    R_gs = auto()

    R_cf = auto() # main cc's
    R_zf = auto()
    R_sf = auto()
    R_of = auto()

    R_pf = auto() # additional cc's
    R_af = auto()
    R_tf = auto()
    R_if = auto()
    R_df = auto()

    R_efl = auto() # eflags

    # the following registers will be used in the disassembly
    # starting from ida v5.7

    R_st0 = auto() # floating point registers(not used in disassembly)
    R_st1 = auto()
    R_st2 = auto()
    R_st3 = auto()
    R_st4 = auto()
    R_st5 = auto()
    R_st6 = auto()
    R_st7 = auto()
    R_fpctrl = auto() # fpu control register
    R_fpstat = auto() # fpu status register
    R_fptags = auto() # fpu tags register

    R_mm0 = auto() # mmx registers
    R_mm1 = auto()
    R_mm2 = auto()
    R_mm3 = auto()
    R_mm4 = auto()
    R_mm5 = auto()
    R_mm6 = auto()
    R_mm7 = auto()

    R_xmm0 = auto() # xmm registers
    R_xmm1 = auto()
    R_xmm2 = auto()
    R_xmm3 = auto()
    R_xmm4 = auto()
    R_xmm5 = auto()
    R_xmm6 = auto()
    R_xmm7 = auto()
    R_xmm8 = auto()
    R_xmm9 = auto()
    R_xmm10 = auto()
    R_xmm11 = auto()
    R_xmm12 = auto()
    R_xmm13 = auto()
    R_xmm14 = auto()
    R_xmm15 = auto()
    R_mxcsr = auto()

    R_ymm0 = auto() # AVX 256 - bit registers
    R_ymm1 = auto()
    R_ymm2 = auto()
    R_ymm3 = auto()
    R_ymm4 = auto()
    R_ymm5 = auto()
    R_ymm6 = auto()
    R_ymm7 = auto()
    R_ymm8 = auto()
    R_ymm9 = auto()
    R_ymm10 = auto()
    R_ymm11 = auto()
    R_ymm12 = auto()
    R_ymm13 = auto()
    R_ymm14 = auto()
    R_ymm15 = auto()

    R_bnd0 = auto() # MPX registers
    R_bnd1 = auto()
    R_bnd2 = auto()
    R_bnd3 = auto()

    R_xmm16 = auto() # AVX - 512 extended XMM registers
    R_xmm17 = auto()
    R_xmm18 = auto()
    R_xmm19 = auto()
    R_xmm20 = auto()
    R_xmm21 = auto()
    R_xmm22 = auto()
    R_xmm23 = auto()
    R_xmm24 = auto()
    R_xmm25 = auto()
    R_xmm26 = auto()
    R_xmm27 = auto()
    R_xmm28 = auto()
    R_xmm29 = auto()
    R_xmm30 = auto()
    R_xmm31 = auto()

    R_ymm16 = auto() # AVX - 512 extended YMM registers
    R_ymm17 = auto()
    R_ymm18 = auto()
    R_ymm19 = auto()
    R_ymm20 = auto()
    R_ymm21 = auto()
    R_ymm22 = auto()
    R_ymm23 = auto()
    R_ymm24 = auto()
    R_ymm25 = auto()
    R_ymm26 = auto()
    R_ymm27 = auto()
    R_ymm28 = auto()
    R_ymm29 = auto()
    R_ymm30 = auto()
    R_ymm31 = auto()

    R_zmm0 = auto() # AVX - 512 ZMM registers
    R_zmm1 = auto()
    R_zmm2 = auto()
    R_zmm3 = auto()
    R_zmm4 = auto()
    R_zmm5 = auto()
    R_zmm6 = auto()
    R_zmm7 = auto()
    R_zmm8 = auto()
    R_zmm9 = auto()
    R_zmm10 = auto()
    R_zmm11 = auto()
    R_zmm12 = auto()
    R_zmm13 = auto()
    R_zmm14 = auto()
    R_zmm15 = auto()
    R_zmm16 = auto()
    R_zmm17 = auto()
    R_zmm18 = auto()
    R_zmm19 = auto()
    R_zmm20 = auto()
    R_zmm21 = auto()
    R_zmm22 = auto()
    R_zmm23 = auto()
    R_zmm24 = auto()
    R_zmm25 = auto()
    R_zmm26 = auto()
    R_zmm27 = auto()
    R_zmm28 = auto()
    R_zmm29 = auto()
    R_zmm30 = auto()
    R_zmm31 = auto()

    R_k0 = auto() # AVX - 512 opmask registers
    R_k1 = auto()
    R_k2 = auto()
    R_k3 = auto()
    R_k4 = auto()
    R_k5 = auto()
    R_k6 = auto()
    R_k7 = auto()

    R_last = auto()


def hasSIB(op: ida_ua.op_t) -> bool:
    """specflag1 indicates if there is a SIB or not"""
    return bool(op.specflag1)


def sib(op: ida_ua.op_t) -> int:
    """specflag2 holds the SIB if there is one"""
    return op.specflag2


def insn_jcc(insn: ida_ua.insn_t) -> bool:
    """Determine if an instruction is a Jcc (jump) instruction"""
    return insn.itype in (
        ida_allins.NN_ja,
        ida_allins.NN_jae,
        ida_allins.NN_jb,
        ida_allins.NN_jbe,
        ida_allins.NN_jc,
        ida_allins.NN_je,
        ida_allins.NN_jg,
        ida_allins.NN_jge,
        ida_allins.NN_jl,
        ida_allins.NN_jle,
        ida_allins.NN_jna,
        ida_allins.NN_jnae,
        ida_allins.NN_jnb,
        ida_allins.NN_jnbe,
        ida_allins.NN_jnc,
        ida_allins.NN_jne,
        ida_allins.NN_jng,
        ida_allins.NN_jnge,
        ida_allins.NN_jnl,
        ida_allins.NN_jnle,
        ida_allins.NN_jno,
        ida_allins.NN_jnp,
        ida_allins.NN_jns,
        ida_allins.NN_jnz,
        ida_allins.NN_jo,
        ida_allins.NN_jp,
        ida_allins.NN_jpe,
        ida_allins.NN_jpo,
        ida_allins.NN_js,
        ida_allins.NN_jz,
    )


def insn_default_opsize_64(insn: ida_ua.insn_t) -> bool:
    """Determine, based on the instruction type, if the instruction, by default, is 64-bit"""
    if insn_jcc(insn):
        return True

    return insn.itype in (
        # use ss
        ida_allins.NN_pop,
        ida_allins.NN_popf,
        ida_allins.NN_popfq,
        ida_allins.NN_push,
        ida_allins.NN_pushf,
        ida_allins.NN_pushfq,
        ida_allins.NN_retn,
        ida_allins.NN_retf,
        ida_allins.NN_retnq,
        ida_allins.NN_retfq,
        ida_allins.NN_call,
        ida_allins.NN_callfi,
        ida_allins.NN_callni,
        ida_allins.NN_enter,
        ida_allins.NN_enterq,
        ida_allins.NN_leave,
        ida_allins.NN_leaveq,
        # near branches
        ida_allins.NN_jcxz,
        ida_allins.NN_jecxz,
        ida_allins.NN_jrcxz,
        ida_allins.NN_jmp,
        ida_allins.NN_jmpni,
        ida_allins.NN_jmpshort,
        ida_allins.NN_loop,
        ida_allins.NN_loopq,
        ida_allins.NN_loope,
        ida_allins.NN_loopqe,
        ida_allins.NN_loopne,
        ida_allins.NN_loopqne,
    )


def mode16(insn: ida_ua.insn_t) -> bool:
    """16-bit mode?"""
    return (insn.auxpref & (aux_use32 | aux_use64)) == 0


def mode32(insn: ida_ua.insn_t) -> bool:
    """32-bit mode?"""
    return (insn.auxpref & aux_use32) != 0


def mode64(insn: ida_ua.insn_t) -> bool:
    """64-bit mode?"""
    return (insn.auxpref & aux_use64) != 0


def natad(insn: ida_ua.insn_t) -> bool:
    """natural address size (no prefixes)?"""
    return (insn.auxpref & aux_natad) != 0


def natop(insn: ida_ua.insn_t) -> bool:
    """natural operand size (no prefixes)?"""
    return (insn.auxpref & aux_natop) != 0


def vexpr(insn: ida_ua.insn_t) -> bool:
    """VEX encoding used"""
    return (insn.auxpref & aux_vexpr) != 0


def evexpr(insn: ida_ua.insn_t) -> bool:
    """EVEX encoding used"""
    return (insn.auxpref & aux_evex) != 0


def xopexpr(insn: ida_ua.insn_t) -> bool:
    """XOP encoding used"""
    return (insn.auxpref & aux_xop) != 0


def ad16(insn: ida_ua.insn_t) -> bool:
    """is current addressing 16-bit?"""
    p = insn.auxpref & (aux_use32 | aux_use64 | aux_natad)
    return p == aux_natad or p == aux_use32


def ad32(insn: ida_ua.insn_t) -> bool:
    """is current addressing 32-bit?"""
    p = insn.auxpref & (aux_use32 | aux_use64 | aux_natad)
    return (
        p == (aux_natad | aux_use32)
        or p == 0
        or p == aux_use64
    )


def ad64(insn: ida_ua.insn_t) -> bool:
    """is current addressing 64-bit?"""
    if not idc.__EA64__:
        return False
    p = insn.auxpref & (aux_use32 | aux_use64 | aux_natad)
    return p == (aux_natad | aux_use64)


def op16(insn: ida_ua.insn_t) -> bool:
    """is current operand size 16-bit?"""
    p = insn.auxpref & (aux_use32 | aux_use64 | aux_natop)
    return (
        p == aux_natop     # 16-bit segment, no prefixes
        or p == aux_use32  # 32-bit segment, 66h
        or p == aux_use64 and (insn.insnpref & REX_W) == 0  # 64-bit segment, 66h, no rex.w
    )


def op32(insn: ida_ua.insn_t) -> bool:
    """is current operand size 32-bit?"""
    p = insn.auxpref & (aux_use32 | aux_use64 | aux_natop)
    return (
        p == 0                           # 16-bit segment, 66h
        or p == (aux_use32 | aux_natop)  # 32-bit segment, no prefiexes
        or p == (aux_use64 | aux_natop) and (insn.insnpref & REX_W) == 0  # 64-bit segment, 66h, no rex.w
    )


def op64(insn: ida_ua.insn_t) -> bool:
    """is current operand size 64-bit?"""
    if not idc.__EA64__:
        return False
    return (
        mode64(insn)
        and (
            (insn.insnpref & REX_W) != 0
            or natop(insn) and insn_default_opsize_64(insn)  # 64-bit segment, rex.w or insns-64
        )
    )


def op256(insn: ida_ua.insn_t) -> bool:
    """is VEX.L == 1 or EVEX.L'L == 01?"""
    return (
        (insn.insnpref & VEX_L) != 0
        and (
            vexpr(insn)
            or xopexpr(insn)
            or evexpr(insn) and (insn.Op6.specflag2 & EVEX_L) == 0
        )
    )


def op512(insn: ida_ua.insn_t) -> bool:
    """is EVEX.L'L == 10?"""
    return (
        evexpr(insn)
        and (insn.insnpref & VEX_L) == 0
        and (insn.Op6.specflag2 & EVEX_L) != 0
    )


def is_vsib(insn: ida_ua.insn_t) -> bool:
    """does instruction use VSIB variant of the sib byte?"""
    return insn.itype in (
        ida_allins.NN_vgatherdps,
        ida_allins.NN_vgatherdpd,
        ida_allins.NN_vgatherqps,
        ida_allins.NN_vgatherqpd,
        ida_allins.NN_vpgatherdd,
        ida_allins.NN_vpgatherdq,
        ida_allins.NN_vpgatherqd,
        ida_allins.NN_vpgatherqq,

        ida_allins.NN_vscatterdps,
        ida_allins.NN_vscatterdpd,
        ida_allins.NN_vscatterqps,
        ida_allins.NN_vscatterqpd,
        ida_allins.NN_vpscatterdd,
        ida_allins.NN_vpscatterdq,
        ida_allins.NN_vpscatterqd,
        ida_allins.NN_vpscatterqq,

        ida_allins.NN_vgatherpf0dps,
        ida_allins.NN_vgatherpf0qps,
        ida_allins.NN_vgatherpf0dpd,
        ida_allins.NN_vgatherpf0qpd,
        ida_allins.NN_vgatherpf1dps,
        ida_allins.NN_vgatherpf1qps,
        ida_allins.NN_vgatherpf1dpd,
        ida_allins.NN_vgatherpf1qpd,

        ida_allins.NN_vscatterpf0dps,
        ida_allins.NN_vscatterpf0qps,
        ida_allins.NN_vscatterpf0dpd,
        ida_allins.NN_vscatterpf0qpd,
        ida_allins.NN_vscatterpf1dps,
        ida_allins.NN_vscatterpf1qps,
        ida_allins.NN_vscatterpf1dpd,
        ida_allins.NN_vscatterpf1qpd,
    )


def vsib_index_fixreg(insn: ida_ua.insn_t, index: regnum_t) -> regnum_t:
    insn_type = insn.itype

    if insn_type in (
        ida_allins.NN_vscatterdps,
        ida_allins.NN_vscatterqps,
        ida_allins.NN_vscatterqpd,
        ida_allins.NN_vpscatterdd,
        ida_allins.NN_vpscatterqd,
        ida_allins.NN_vpscatterqq,

        ida_allins.NN_vpgatherdd,
        ida_allins.NN_vpgatherqd,
        ida_allins.NN_vpgatherqq,
        ida_allins.NN_vgatherdps,
        ida_allins.NN_vgatherqps,
        ida_allins.NN_vgatherqpd,
    ):
        if index > 15:
            if op512(insn):
                index += RegNo.R_zmm0
            elif op256(insn):
                index += RegNo.R_ymm16 - 16
            else:
                index += RegNo.R_xmm16 - 16
        else:
            if op512(insn):
                index += RegNo.R_zmm0
            elif op256(insn):
                index += RegNo.R_ymm0
            else:
                index += RegNo.R_xmm0

    elif insn_type in (
        ida_allins.NN_vscatterdpd,
        ida_allins.NN_vpscatterdq,

        ida_allins.NN_vgatherdpd,
        ida_allins.NN_vpgatherdq,
    ):
        if index > 15:
            if op512(insn):
                index += RegNo.R_ymm16 - 16
            else:
                index += RegNo.R_xmm16 - 16
        else:
            if op512(insn):
                index += RegNo.R_ymm0
            else:
                index += RegNo.R_xmm0

    elif insn_type in (
        ida_allins.NN_vgatherpf0dps,
        ida_allins.NN_vgatherpf0qps,
        ida_allins.NN_vgatherpf0qpd,
        ida_allins.NN_vgatherpf1dps,
        ida_allins.NN_vgatherpf1qps,
        ida_allins.NN_vgatherpf1qpd,

        ida_allins.NN_vscatterpf0dps,
        ida_allins.NN_vscatterpf0qps,
        ida_allins.NN_vscatterpf0qpd,
        ida_allins.NN_vscatterpf1dps,
        ida_allins.NN_vscatterpf1qps,
        ida_allins.NN_vscatterpf1qpd,
    ):
      index += RegNo.R_zmm0

    elif insn_type in (
        ida_allins.NN_vgatherpf0dpd,
        ida_allins.NN_vgatherpf1dpd,
        ida_allins.NN_vscatterpf0dpd,
        ida_allins.NN_vscatterpf1dpd,
    ):
        if index > 15:
            index += RegNo.R_ymm16 - 16
        else:
            index += RegNo.R_ymm0

    return index


def sib_base(insn: ida_ua.insn_t, op: ida_ua.op_t) -> regnum_t:
    """Calculate the base register number for a phrase/displacment"""
    base = sib(op) & 7
    if idc.__EA64__ and (insn.insnpref & REX_B):  # Do we need to convert the base to a 64-bit register number?
        base |= 8  # Upconvert to 64-bit register number if not already

    return base


def sib_index(insn: ida_ua.insn_t, op: ida_ua.op_t) -> regnum_t:
    """Calculate the index register number for a phrase/displacement"""
    index = (sib(op) >> 3) & 7
    if idc.__EA64__ and (insn.insnpref & REX_X):  # Do we need to conver the index to a 64-bit register number?
        index |= 8  # Upconvert to 64-bit register number if not already
    if is_vsib(insn):
        # Op6 is used for opmask registers in EVEX.
        # spec flags from Op6 are used to extend insn_t.
        if (insn.Op6.specflag2 & EVEX_V) != 0:
            index |= 16
        index = vsib_index_fixreg(insn, index)

    return index


def sib_scale(op: ida_ua.op_t) -> int:
    """Calculate the scale for the index register"""
    return (sib(op) >> 6) & 3


def x86_base_reg(insn: ida_ua.insn_t, op: ida_ua.op_t) -> regnum_t:
    """
    Get the base register of the operand with a displacement.
    Returns correct register for 16-bit code too
    """
    if hasSIB(op):
        if op.type == ida_ua.o_mem:
            return RegNo.R_none
        return sib_base(insn, op)  # base register encoded in the SIB

    if not ad16(insn):
        return op.phrase  # "phrase" contains the base register number

    if signed(op.phrase, get_bits()) == RegNo.R_none:
        return RegNo.R_sp

    if op.phrase in (0, 1, 7):  # ([BX+SI], [BX+DI], [BX])
        return RegNo.R_bx

    if op.phrase in (2, 3, 6):  # ([BP+SI], [BP+DI], [BP])
        return RegNo.R_bp

    if op.phrase == 4:  # [SI]
        return RegNo.R_si

    if op.phrase == 5:  # [DI]
        return RegNo.R_di

    raise ValueError("Unable to parse x86 base register from instruction")


INDEX_NONE = 4  # no index register is present


def x86_index_reg(insn: ida_ua.insn_t, op: ida_ua.op_t) -> regnum_t:
    """Get the index register (if there is one) (handle 16-bit as well for completeness)"""
    if hasSIB(op):
        idx = sib_index(insn, op)
        if idx != INDEX_NONE:
            return idx
        return RegNo.R_none

    if not ad16(insn):
        return RegNo.R_none

    if op.phrase in (0, 2):  # ([BX+SI], [BP+SI])
        return RegNo.R_si

    if op.phrase in (1, 3):  # ([BX+DI], [BP+DI])
        return RegNo.R_di

    if op.phrase in (4, 5, 6, 7):  # ([SI], [DI], [BP], [BX])
        return RegNo.R_none

    raise ValueError("Unable to parse x86 index register from instruction")


def x86_scale(op: ida_ua.op_t) -> int:
    """get the scale factor of the operand with a displacement"""
    return sib_scale(op) if hasSIB(op) else 0


def has_displ(op: ida_ua.op_t) -> bool:
    """does the operand have a displacement?"""
    return op.type == ida_ua.o_displ or op.type == ida_ua.o_mem and hasSIB(op)



"""
CPU Emulator Built-in Windows API functions

These functions are used to emulate the effects of "useful" Windows API functions.

Add any builtin functions that need to be handled below.  The function should be declared as such

# Using the same function for multiple API functions
@builtin_func("GetSystemDirectoryA")
@builtin_func("GetSystemDirectoryW")
#typespec("UINT GetSystemDirectoryA(char* lpBuffer, UINT uSize);")
def memmove(cpu_context, func_name, func_args):
    logger.debug("Emulating GetSystemDirectoryA")

"""
import logging
import ntpath

from ... import constants
from ...call_hooks import builtin_func


logger = logging.getLogger(__name__)


@builtin_func("PathAppendA")
@builtin_func("PathAppendW")
#typedef(BOOL PathAppendA(LPSTR  pszPath,LPCSTR pszMore);)
def pathappend(cpu_context, func_name, func_args):
    """
    Appends one path to the end of another
    """
    wide = func_name.endswith(u"W")
    path_ptr, more_ptr = func_args

    curr_path = cpu_context.read_data(
        path_ptr, data_type=constants.WIDE_STRING if wide else constants.STRING
    ).decode("utf-16-le" if wide else "utf8")
    more_path = cpu_context.read_data(
        more_ptr, data_type=constants.WIDE_STRING if wide else constants.STRING
    ).decode("utf-16-le" if wide else "utf8")

    full_path = ntpath.join(curr_path, more_path)
    cpu_context.write_data(path_ptr, full_path, data_type=constants.WIDE_STRING if wide else constants.STRING)
    return True


@builtin_func("PathAddBackslashA")
@builtin_func("PathAddBackslashW")
#typedef(LPWSTR PathAddBackslashW(LPWSTR pszPath));)
def pathaddbackslash(cpu_context, func_name, func_args):
    """
    Appends a backslash to the path
    """
    wide = func_name.endswith(u"W")
    path_ptr = func_args[0]

    curr_path = cpu_context.read_data(
        path_ptr, data_type=constants.WIDE_STRING if wide else constants.STRING
    ).decode("utf-16-le" if wide else "utf8")

    full_path = curr_path + "\\"
    cpu_context.write_data(path_ptr, full_path, data_type=constants.WIDE_STRING if wide else constants.STRING)
    return True



# Expose kordesii API
from .core import *
from .logutil import setup_logging, get_logger
from .serialization import get_serializer
from .reporter import Reporter
from .tester import Tester
from .registry import (
    register_entry_points, register_decoder_directory, register_decoder_package,
    iter_decoders, get_decoder_descriptions, set_default_source, clear_default_source,
    get_sources)
from .ida_client import IDA, run_in_ida

# Add do nothing logger for when Kordesii is used as a library.
import logging

logging.getLogger('kordesii').addHandler(logging.NullHandler())



"""
Components for emulating an ARM architecture.
"""
from .instruction import ARMInstruction
from ..cpu_context import ProcessorContext
from .registers import ARM_Registers
from .opcodes import OPCODES
from ..utils import get_bits


STACK_BASE = 0x1180000  # Base address for stack

SP_OFFSET = 0x800
FP_OFFSET = 0x400


class ARMProcessorContext(ProcessorContext):
    """Processor context for ARM architecture"""

    OPCODES = OPCODES.copy()
    _instruction_class = ARMInstruction

    def __init__(self, emulator):
        super().__init__(
            emulator,
            ARM_Registers(),
            instruction_pointer="pc",
            stack_pointer="sp",
            return_register="x0",  # TODO: Specify r0 or x0 based on bitness?
        )
        # Set up the stack before we go.
        self.registers.sp = STACK_BASE - SP_OFFSET
        if get_bits() == 64:
            self.registers.x29 = STACK_BASE - FP_OFFSET
        else:
            self.registers.r11 = STACK_BASE - FP_OFFSET



"""
CPU Emulator Built-in Windows API functions

These functions are used to emulate the effects of "useful" Windows API functions.

Add any builtin functions that need to be handled below.  The function should be declared as such

# Using the same function for multiple API functions
@builtin_func("GetSystemDirectoryA")
@builtin_func("GetSystemDirectoryW")
#typespec("UINT GetSystemDirectoryA(char* lpBuffer, UINT uSize);")
def memmove(cpu_context, func_name, func_args):
    logger.debug("Emulating GetSystemDirectoryA")

"""

import time
import random
import logging

import ida_nalt

from . import win_constants as wc
from ...call_hooks import builtin_func
from ... import constants
from ... import actions, objects

logger = logging.getLogger(__name__)


# Maps environment variables to the most common values.
ENV_VAR_MAP = {
    "comspec": "C:\\Windows\\system32\\cmd.exe",
    "windir": "C:\\Windows",
    "systemroot": "C:\\Windows",
    "systemdrive": "C:",
    "homedrive": "C:",
}


@builtin_func("GetEnvironmentVariableA")
@builtin_func("GetEnvironmentVariableW")
#typespec("DWORD GetEnvironmentVariableA(LPCSTR lpName, LPSTR lpBuffer, DWORD nSize)")
def get_environment_variable(cpu_context, func_name, func_args):
    """
    Retrieves the contents of the specified variable from the environment block of the calling process.
    """
    wide = func_name.endswith("W")
    var_name_ptr, buffer_ptr, max_size = func_args
    var_name = cpu_context.read_data(
        var_name_ptr, data_type=constants.WIDE_STRING if wide else constants.STRING
    ).decode("utf-16-le" if wide else "utf8")

    # Replace some common environment variables, and add %'s for others.
    # - 1 for null terminator
    name = ENV_VAR_MAP.get(var_name.lower(), f"%{var_name}%")[:max_size - 1]
    cpu_context.write_data(buffer_ptr, name, data_type=constants.WIDE_STRING if wide else constants.STRING)
    logger.debug("Getting environment variable: %s -> %s", var_name, name)

    return len(name)


@builtin_func("GetCurrentDirectoryA")
@builtin_func("GetCurrentDirectoryW")
#typespec("DWORD GetCurrentDirectory(DWORD  nBufferLength, LPTSTR lpBuffer);")
def get_current_directory(cpu_context, func_name, func_args):
    """
    Retrieves the current working directory

    Using value "." to represent the current working directory for this emulator
    """
    wide = func_name.endswith("W")
    max_size, buffer_ptr = func_args

    cwd = u"."[:max_size]

    logger.debug("Writing current working directory %s to 0x%08X", cwd, buffer_ptr)
    cpu_context.write_data(buffer_ptr, cwd, data_type=constants.WIDE_STRING if wide else constants.STRING)

    return len(cwd)


@builtin_func("GetModuleFileNameA")
@builtin_func("GetModuleFileNameW")
#typespec("DWORD GetModuleFileNameA_1(HMODULE hModule, LPSTR lpFilename, DWORD nSize);")
def get_module_file_name(cpu_context, func_name, func_args):
    r"""
    Get the fully qualified path for the file that contains the specified module.

    Using the real filename prefixed by "%INPUT_FILE_DIR%\" to indicate the file path for this emulator.
    """
    wide = func_name.endswith("W")
    module_handle, filename_ptr, max_size = func_args

    # We don't support getting filename for modules that aren't itself.
    if module_handle != 0:
        return

    # Getting input file path as module path.
    # Since we shouldn't expose the user's real file path structure we'll use %INPUT_FILE_DIR% instead.
    # Must be truncated to fit in max_size filename_ptr (-1 for the terminator)
    file_path = "%INPUT_FILE_DIR%\\" + ida_nalt.get_root_filename()
    file_path = file_path[:max_size - 1]

    logger.debug("Writing module path %s to 0x%08X", file_path, filename_ptr)
    cpu_context.write_data(
        filename_ptr,
        file_path,
        data_type=constants.WIDE_STRING if wide else constants.STRING,
    )

    return len(file_path)


@builtin_func("GetShortPathNameA")
@builtin_func("GetShortPathNameW")
#typespec("DWORD GetShortPathNameA(LPCSTR lpszLongPath, LPSTR lpszShortPath, DWORD cchBuffer)")
def get_short_path_name(cpu_context, func_name, func_args):
    """
    Retrieves the short path form of the specified path.
    """
    # We aren't going to bother actually shortening the path name.
    # This hook is just here to ensure the destination gets filled in.
    wide = func_name.endswith("W")
    long_path_ptr, short_path_ptr, short_path_size = func_args

    long_path = cpu_context.read_data(
        long_path_ptr, data_type=constants.WIDE_STRING if wide else constants.STRING
    ).decode("utf-16-le" if wide else "utf8")

    short_path = long_path[:short_path_size]
    cpu_context.write_data(
        short_path_ptr, short_path, data_type=constants.WIDE_STRING if wide else constants.STRING
    )
    logger.debug("Copying %r from 0x%08X -> 0x%08X", short_path, long_path_ptr, short_path_ptr)

    return len(short_path)


@builtin_func("GetSystemDirectoryA")
@builtin_func("GetSystemDirectoryW")
#typespec("UINT GetSystemDirectoryA(char* lpBuffer, UINT uSize);")
#typespec("UINT GetSystemDirectoryW(wchar* lpbuffer, UINT uSize);")
def get_system_directory(cpu_context, func_name, func_args):
    r"""
    Retrieves the path of the system directory.

    Using unexpanded "%WinDir%\System32" to indicate the system directory for this emulator.
    """
    wide = func_name.endswith("W")
    buffer_ptr, max_size = func_args

    system_dir = r"%WinDir%\System32"[:max_size]

    logger.debug("Writing system directory %s to 0x%08X", system_dir, buffer_ptr)
    cpu_context.write_data(
        buffer_ptr,
        system_dir,
        data_type=constants.WIDE_STRING if wide else constants.STRING,
    )

    return len(system_dir)


@builtin_func("GetTempPathA")
@builtin_func("GetTempPathW")
#typespec("DWORD GetTempPathA(DWORD nBufferLength,LPSTR lpBuffer);")
def get_temp_path(cpu_context, func_name, func_args):
    """
    Retrieves the path of the temp directory

    Using unexpanded "%Temp%\" to indicate the temp directory for this emulator
    """
    wide = func_name.endswith("W")
    max_size, buffer_ptr = func_args

    # The returned string ends with a backslash
    temp_dir = u"%Temp%\\"[:max_size]

    logger.debug("Writing temp directory %s to 0x%08X}", temp_dir, buffer_ptr)
    cpu_context.write_data(buffer_ptr, temp_dir, data_type=constants.WIDE_STRING if wide else constants.STRING)

    return len(temp_dir)


@builtin_func("GetTickCount")
@builtin_func("GetTickCount64")
#typespec("DWORD GetTickCount();")
#typespec("ULONGLONG GetTickCount64();")
def get_tick_count(cpu_context, func_name, func_args):
    """
    Retrieves the number of milliseconds that have elapsed since the system was started, up to 49.7 days.
    """
    # TODO: Should we allow users to specify a specific tick count or provide permission
    #   to use the real time when setting up Emulator?
    # mask = 0xFFFFFFFFFFFFFFFF if func_name == "GetTickCount64" else 0xFFFFFFFF
    # return int(time.time()) & mask

    # Returning a fake tick count in order for results to be more deterministic.
    return 1587410693


@builtin_func("GetWindowsDirectoryA")
@builtin_func("GetWindowsDirectoryW")
#typespec("UINT GetWindowsDirectoryA(char* lpBuffer, UINT uSize);")
#typespec("UINT GetWindowsDirectoryW(wchar* lpBuffer, UINT uSize);")
def get_windows_directory(cpu_context, func_name, func_args):
    """
    Retrieves the path of the Windows directory.

    Using unexpanded "%WinDir%" to indicate the windows directory for this emulator.
    """
    wide = func_name.endswith("W")
    buffer_ptr, max_size = func_args

    win_dir = "%WinDir%"[:max_size]
    logger.debug("Writing windows directory %s to 0x%08X", win_dir, buffer_ptr)
    cpu_context.write_data(
        buffer_ptr,
        win_dir,
        data_type=constants.WIDE_STRING if wide else constants.STRING,
    )

    return len(win_dir)


@builtin_func("CreateNamedPipeA")
#typespec("HANDLE CreateNamedPipeA(char* lpName, DWORD dwOpenMode, DWORD dwPipeMode, DWORD nMaxInstances, DWORD nOutBufferSize, DWORD nInBufferSize, DWORD nDefaultTimeOut, SECURITY_ATTRIBUTES* lpSecurityAttributes);")
def create_named_pipe(cpu_context, func_name, func_args):
    """
    Creates an instance of a named pipe and returns a handle for subsequent pip operations.
    """
    # TODO: Create high level NamedPipe objects to keep track of data stored in pipes.
    #   This should work similar if not the same to the File objects we already keep track of.
    return random.randint(wc.MIN_HANDLE, wc.MAX_HANDLE)


@builtin_func("GetComputerNameA")
@builtin_func("GetComputerNameW")
#typespec("BOOL GetComputerNameA(char* lpBuffer, DWORD* nSize);")
#typespec("BOOL GetComputerNameW(wchar* lpBuffer, DWORD* nSize);")
def get_computer_name(cpu_context, func_name, func_args):
    """
    Retrieves the NetBIOS name of the local computer.
    This name is established at system startup, when the system reads it
    from the registry.

    Using the computer name "KORDESII_COMP" for this emulator.
    TODO: Should we allow user to provide their own computer name when setting up Emulator?
    """
    wide = func_name.endswith("W")
    buffer_ptr, size_ptr = func_args

    computer_name = "KORDESII_COMP"
    logger.debug("Writing computer name %s to 0x%08X", computer_name, buffer_ptr)
    cpu_context.write_data(
        buffer_ptr,
        computer_name,
        data_type=constants.WIDE_STRING if wide else constants.STRING
    )
    cpu_context.write_data(size_ptr, len(computer_name), data_type=constants.DWORD)

    return 1  # return success


@builtin_func("GetUserNameA")
@builtin_func("GetUserNameW")
#typespec("BOOL GetUserNameA(char* lpBuffer, DWORD* pcbBuffer);")
#typespec("BOOL GetUserNameW(wchar* lpBuffer, DWORD* pcbBuffer);")
def get_user_name(cpu_context, func_name, func_args):
    """
    Retrieves the name of the user associated with the current thread.

    Using the user name "kordesii" for this emulator.
    TODO: Should we allow user to provide their own user name when setting up Emulator?
    """
    wide = func_name.endswith("W")
    buffer_ptr, size_ptr = func_args

    user_name = "kordesii"
    logger.debug("Writing user name %s to 0x%08X", user_name, buffer_ptr)
    cpu_context.write_data(
        buffer_ptr,
        user_name,
        data_type=constants.WIDE_STRING if wide else constants.STRING,
    )
    cpu_context.write_data(size_ptr, len(user_name), data_type=constants.DWORD)

    return 1  # return success


@builtin_func("CreateProcessA")
@builtin_func("CreateProcessW")
#typespec("BOOL CreateProcessA(LPCSTR lpApplicationName, LPSTR lpCommandLine, LPSECURITY_ATTRIBUTES lpProcessAttributes, LPSECURITY_ATTRIBUTES lpThreadAttributes, BOOL bInheritHandles, DWORD dwCreationFlags, LPVOID lpEnvironment, LPCSTR lpCurrentDirectory, LPSTARTUPINFOA lpStartupInfo, LPPROCESS_INFORMATION lpProcessInformation)")
def create_process(cpu_context, func_name, func_args):
    """
    Create a new process.
    """
    wide = func_name.endswith("W")

    app_ptr = func_args[0]
    cmd_ptr = func_args[1]

    cmd = cpu_context.read_data(
        cmd_ptr, data_type=constants.WIDE_STRING if wide else constants.STRING
    ).decode("utf-16-le" if wide else "utf8")
    if app_ptr:
        app = cpu_context.read_data(
            app_ptr, data_type=constants.WIDE_STRING if wide else constants.STRING
        ).decode("utf-16-le" if wide else "utf8")
        cmd = app + " " + cmd

    logger.debug("%s: %r", func_name, cmd)
    cpu_context.actions.add(actions.CommandExecuted(cpu_context.ip, cmd))

    return random.randint(wc.MIN_HANDLE, wc.MAX_HANDLE)


@builtin_func("WinExec")
#typespec("UINT WinExec(char* lpCmdLine, UINT nCmdShow);")
def win_exec(cpu_context, func_name, func_args):
    """
    Runs the specified application.
    """
    cmd_ptr, visibility = func_args
    cmd = cpu_context.read_data(cmd_ptr).decode("utf8")

    logger.debug("WinExec: %r", cmd)
    cpu_context.actions.add(
        actions.CommandExecuted(cpu_context.ip, cmd, wc.Visibility(visibility))
    )

    return random.randint(wc.MIN_HANDLE, wc.MAX_HANDLE)


@builtin_func("CreateDirectoryA")
@builtin_func("CreateDirectoryW")
#typespec("BOOL CreateDirectoryA(char* lpPathName, SECURITY_ATTRIBUTES* lpSecurityAttributes);")
#typespec("BOOL CreateDirectoryw(wchar* lpPathName, SECURITY_ATTRIBUTES* lpSecurityAttributes);")
def create_directory(cpu_context, func_name, func_args):
    """
    Creates a new directory.
    """
    wide = func_name.endswith("W")
    path_ptr = func_args[0]

    path = cpu_context.read_data(path_ptr, data_type=constants.WIDE_STRING if wide else constants.STRING)
    path = path.decode("utf-16-le" if wide else "utf8")
    logger.debug("Create Directory: %r", path)

    if path:
        cpu_context.actions.add(actions.DirectoryCreated(cpu_context.ip, path))

    return 1  # return success


@builtin_func("CreateFileA")
@builtin_func("CreateFileW")
#typespec("HANDLE CreateFileA(char* lpFileName, DWORD dwDesiredAccess, DWORD dwShareMode, SECURITY_ATTRIBUTES* lpSecurityAttributes, DWORD dwCreationDisposition, DWORD dwFlagsAndAttributes, HANDLE hTemplateFile);")
#typespec("HANDLE CreateFileW(wchar* lpFileName, DWORD dwDesiredAccess, DWORD dwShareMode, SECURITY_ATTRIBUTES* lpSecurityAttributes, DWORD dwCreationDisposition, DWORD dwFlagsAndAttributes, HANDLE hTemplateFile);")
def create_file(cpu_context, func_name, func_args):
    """
    Creates or opens a file or I/O device.
    The function returns a handle that can be used to access the file or device for various
    types of I/O depending on the file or device and the flags and attributes specified.
    """
    wide = func_name.endswith("W")
    name_ptr = func_args[0]

    path = cpu_context.read_data(name_ptr, data_type=constants.WIDE_STRING if wide else constants.STRING)
    path = path.decode("utf-16-le" if wide else "utf8")

    mode = ""
    desired_access = func_args[1]
    if desired_access & wc.GENERIC_READ:
        mode += "r"
    if desired_access & wc.GENERIC_WRITE:
        mode += "w"

    if not path:
        path = f"0x{cpu_context.ip:08x}.bin"

    # Even if we have seen the path before, CreateFile* always creates a new
    # handle, so we should too.
    handle = cpu_context.objects.alloc()

    create_disposition = func_args[4]
    if create_disposition in [wc.CREATE_NEW, wc.CREATE_ALWAYS]:
        logger.debug("Created file: %s", path)
        cpu_context.actions.add(actions.FileCreated(cpu_context.ip, handle, path, mode))
    elif create_disposition in [wc.OPEN_EXISTING, wc.OPEN_ALWAYS]:
        logger.debug("Opened file: %s", path)
        cpu_context.actions.add(actions.FileOpened(cpu_context.ip, handle, path, mode))
    elif create_disposition == wc.TRUNCATE_EXISTING:
        logger.debug("Truncated file: %s", path)
        cpu_context.actions.add(actions.FileTruncated(cpu_context.ip, handle, path, mode))
    
    return handle


@builtin_func("WriteFile")
#typespec("BOOL WriteFile(HANDLE hFile, LPCVOID lpBuffer, DWORD nNumberOfBytesToWrite, LPDWORD lpNumberOfBytesWritten, LPOVERLAPPED lpOverlapped)")
def write_file(cpu_context, func_name, func_args):
    """
    Writes data to the specified file.
    """
    handle = func_args[0]
    data_ptr = func_args[1]
    num_bytes_to_write = func_args[2]
    bytes_written_ptr = func_args[3]

    if not handle:
        logger.warning(f"File with handle {hex(handle)} was never opened.")
        return 1

    logger.debug(f"Writing {num_bytes_to_write} bytes to file with handle {hex(handle)}.")
    data = cpu_context.mem_read(data_ptr, num_bytes_to_write)
    cpu_context.actions.add(actions.FileWritten(cpu_context.ip, handle, data))

    if bytes_written_ptr:
        cpu_context.write_data(bytes_written_ptr, len(data), data_type=constants.DWORD)

    return 1  # return success


@builtin_func("MoveFileA")
@builtin_func("MoveFileExA")
def move_file(cpu_context, func_name, func_args):
    """
    Moves an existing file (or directory) to new location.
    """
    old_name_ptr, new_name_ptr, *_ = func_args
    old_path = cpu_context.read_data(old_name_ptr).decode("utf8")
    new_path = cpu_context.read_data(new_name_ptr).decode("utf8")
    logger.debug("Moving file: %s -> %s", old_path, new_path)

    handle = cpu_context.objects.get_or_alloc(objects.File, path=old_path)

    if not new_path:
        logger.debug("Ignoring attempt to change file path to empty string.")
    
    cpu_context.actions.add(actions.FileMoved(cpu_context.ip, handle, old_path, new_path))

    return 1  # return success


@builtin_func("CloseHandle")
#typespec("BOOL __stdcall CloseHandle(HANDLE hObject)")
def close_file(cpu_context, func_name, func_args):
    """
    Closes an open object handle.
    """
    handle = func_args[0]

    if not handle:
        logger.warning(f"File with handle {hex(handle)} does not exist.")
        return 1
        
    logger.debug(f"Closing file with handle {hex(handle)}.")
    cpu_context.actions.add(actions.FileClosed(cpu_context.ip, handle))

    return 1  # return success


@builtin_func("DeleteFileA")
@builtin_func("DeleteFileW")
#typespec("BOOL DeleteFileA(char* lpFileName);")
#typespec("BOOL DeleteFileW(wchar* lpFileName);")
def delete_file(cpu_context, func_name, func_args):
    """
    Deletes an existing file.
    """
    wide = func_name.endswith("W")
    path_ptr = func_args[0]
    path = cpu_context.read_data(path_ptr, data_type=constants.WIDE_STRING if wide else constants.STRING)
    path = path.decode("utf16le" if wide else "utf8")
    logger.debug("Deleting: %s", path)

    handle = cpu_context.objects.get_or_alloc(objects.File, path=path)
    cpu_context.actions.add(actions.FileDeleted(cpu_context.ip, handle, path))

    return 1  # return success


@builtin_func("CreateMutexA")
@builtin_func("CreateMutexW")
@builtin_func("CreateMutexExA")
@builtin_func("CreateMutexExW")
#typespec("HANDLE CreateMutexA(SECURITY_ATTRIBUTES* lpMutexAttributes, BOOL bInitialOwner, char* lpName);")
#typespec("HANDLE CreateMutexW(SECURITY_ATTRIBUTES* lpMutexAttributes, BOOL bInitialOwner, wchar* lpName);")
#typespec("HANDLE CreateMutexExA(SECURITY_ATTRIBUTES* lpMutexAttributes, char* lpName, DWROD dwFlags, DWORD dwDesiredAccess);")
#typespec("HANDLE CreateMutexExW(SECURITY_ATTRIBUTES* lpMutexAttributes, wchar* lpName, DWORD dwFlags, DWORD dwDesiredAccess);")
def create_mutex(cpu_context, func_name, func_args):
    """
    Creates or opens a named or unnamed mutex object and returns a handle to the object.
    """
    # TODO: Create high level Mutex object.
    return random.randint(wc.MIN_HANDLE, wc.MAX_HANDLE)


@builtin_func("CreateEventA")
@builtin_func("CreateEventW")
@builtin_func("CreateEventExA")
@builtin_func("CreateEventEXW")
#typespec("HANDLE CreateEventA(SECURTIY_ATTRIBUTES* lpEventAttributes, BOOL bManualReset, BOOL bInitialState, char* lpName);")
#typespec("HANDLE CreateEventW(SECURTIY_ATTRIBUTES* lpEventAttributes, BOOL bManualReset, BOOL bInitialState, wchar* lpName);")
#typespec("HANDLE CreateEventExA(SECURITY_ATTRIBUTES* lpEventAttributes, char* lpName, DWORD dwFlags, DWORD dwDesiredAccess);")
#typespec("HANDLE CreateEventExW(SECURITY_ATTRIBUTES* lpEventAttributes, wchar* lpName, DWORD dwFlags, DWORD dwDesiredAccess);")
def create_event(cpu_context, func_name, func_args):
    """
    Creates or opens a named or unnamed event object and returns a handle to the object.
    """
    # TODO: Create high level Event object.
    return random.randint(wc.MIN_HANDLE, wc.MAX_HANDLE)


@builtin_func("CreateSemaphoreA")
@builtin_func("CreateSemaphoreW")
@builtin_func("CreateSemaphoreExA")
@builtin_func("CreateSemaphoreExW")
#typespec("HANDLE CreateSemaphoreA(SECURITY_ATTRIBUTES* lpSemaphoreAttributes, LONG lInitialCount, LONG lMaximumCount, char* lpName);")
#typespec("HANDLE CreateSemaphoreW(SECURITY_ATTRIBUTES* lpSemaphoreAttributes, LONG lInitialCount, LONG lMaximumCount, wchar* lpName);")
#typespec("HANDLE CreateSemaphoreExA(SECURITY_ATTRIBUTES* lpSemaphoreAttributes, LONG lInitialCount, LONG lMaximumCount, char* lpName, DWORD dwFlags, DWORD dwDesiredAccess);")
#typespec("HANDLE CreateSemaphoreExW(SECURITY_ATTRIBUTES* lpSemaphoreAttributes, LONG lInitialCount, LONG lMaximumCount, wchar* lpName, DWORD dwFlags, DWORD dwDesiredAccess);")
def create_semaphore(cpu_context, func_name, func_args):
    """
    Creates or opens a named or unnamed semaphore object and returns a handle to the object.
    """
    # TODO: Create high level Semaphore object.
    return random.randint(wc.MIN_HANDLE, wc.MAX_HANDLE)



"""
Utilities for working with encoded strings.
"""
import copy
import functools
import logging
import re
import sys
from typing import List, Iterable, Tuple

import ida_name
import idaapi
import idc
import idautils

import kordesii
from kordesii.serialization import serializable_class
from kordesii.utils import function_tracing

logger = logging.getLogger(__name__)

# fmt: off
# Codecs used to detect encoding of strings.
CODE_PAGES = [
    "ascii",
    "utf-32-be", "utf-32-le", "utf-16-be", "utf-16-le", "utf-8",  # General (utf-7 omitted)
    "gb18030", "gbk",  # Unified Chinese
    "gb2312", "hz",  # Simplified Chinese
    "big5hkscs", "big5",  # Traditional Chinese (cp950 omitted)
    "koi8-r", "iso8859-5", "cp1251", "mac-cyrillic",  # Cyrillic (cp866, cp855 omitted)
    "cp949",  # Korean (johab, iso2022-kr omitted)
    "iso8859-6", "cp1256",  # Arabic (cp864, cp720 omitted)
    "latin1",  # If all else fails, latin1 is always is successful.
]
# fmt: on

# CODEC to use for displaying strings in IDA, etc.
DISPLAY_CODE = "cp437" if sys.platform == "win32" else "ascii"


# TODO: Add support for forcing only utf-16 or utf-8 string data?
def find_string_data(data: bytes) -> Iterable[Tuple[int, bytes, str]]:
    """
    Iterates string data found within the given data.
    This looks for the start of UTF-8 or UTF-16 strings found within the given block of data.

    :param data: Chunk of data that contains a mixture of UTF-8 and UTF-16 encoded strings.
    :yields: (start_offset, string_data, encoding)
    """
    if not data or data in (b"\x00", b"\x00\x00"):
        return

    # Iterate strings split up by either \x00\x00 or \x00.
    # If we find a single character followed by a \x00, this is most likely a character of utf-16.
    offset = 0
    buffer = b""
    utf_16 = False
    for entry in re.split(b"((?<=\x00[^\x00]\x00)\x00\x00|\x00)", data):
        buffer += entry
        offset += len(entry)
        string_offset = offset - len(buffer)

        # If we have a single character, then this is most likely part of a utf-16 string.
        # Don't yield anything yet.
        if len(entry) == 1 and entry != b"\x00":
            utf_16 = True

        # If we find a single null character and not currently extracting a utf-16 string,
        # we found the end of a utf-8 string.
        elif entry == b"\x00" and not utf_16:
            if buffer not in (b"\x00", b"\x00\x00"):
                yield string_offset, buffer, "utf-8"
            buffer = b""

        # If we find a double null character, this is the end of a utf-16 string.
        elif entry == b"\x00\x00":
            if buffer not in (b"\x00", b"\x00\x00"):
                yield string_offset, buffer, "utf-16-le"
            buffer = b""
            utf_16 = False

        # If we find a multi-character string but in utf_16 mode, then we misidentified the utf-16.
        # Yield buffer as single character utf-8 strings and then reset buffer to entry.
        elif len(entry) > 1 and utf_16:
            split_strings = []
            for _entry in re.split(b"(\x00)", buffer[:-len(entry)]):
                if _entry == b"\x00":
                    split_strings[-1] += _entry
                else:
                    split_strings.append(_entry)
            for _entry in split_strings:
                if _entry:
                    yield string_offset, _entry, "utf-8"
                string_offset += len(_entry)
            buffer = buffer[-len(entry):]
            utf_16 = False

    # Yield remaining string that may not have a null terminator.
    if buffer and buffer not in (b"\x00", b"\x00\x00"):
        string_offset = offset - len(buffer)
        if utf_16:
            # If we have no ending null byte and there is a null byte in front, then it must be big endian.
            if not buffer.endswith(b"\x00") and len(buffer) < len(data) and data[string_offset-1] == 0x00:
                buffer = b"\x00" + buffer
                yield string_offset - 1, buffer, "utf-16-be"
            else:
                yield string_offset, buffer, "utf-16-le"
        else:
            yield string_offset, buffer, "utf-8"


@serializable_class
@functools.total_ordering
class EncodedString(object):
    """
    Object to hold data about an encoded/encrypted string.
    """

    _MAX_COMMENT_LENGTH = 130
    _MAX_NAME_LENGTH = 30

    # TODO: Since we must have either a string_location or encoded_data should we change
    #   the first parameter to be something like "encoded_data_or_location" and then
    #   check which it is by type?
    def __init__(
        self,
        string_location,
        string_reference=None,
        size=None,
        offset=None,
        key=None,
        encoded_data=None,
        code_page=None,
        dest=None,
    ):
        """
        Instantiate an EncodedString.

        :param string_location:
            Data segment pointer for static strings or stack pointer for stack strings.
        :param string_reference:
            The location the string is referenced from.
            This is required to pull the stack frame when string_location is a stack pointer.
        :param size: Size of the encoded data.
            When set encoded_data is populated using this and string_location.
        :param offset: Used when there is an offset based accessing scheme.
        :param key: Used when there is a key that can vary by string.
        :param encoded_data: encoded/encrypted data, if not provided data will be retrieved from IDA.
        :param code_page: known encoding page used to decode data to unicode (after data is decrypted)
            (code page is dynamically determined if not provided)
        :param dest: Location of decrypted data (if different from string_location)

        :raises ValueError: If both encoded_data and string_location is not provided.
        """
        self.string_location = string_location
        self.string_reference = string_reference
        self.offset = offset
        self.key = key
        self.decoded_data = None
        self.code_page = None
        self.dest = dest
        self._xrefs_to = None

        # Pull encoded_data from IDA if not provided.
        if encoded_data is None:
            if string_location is None:
                raise ValueError(
                    "encoded_data must be provided if string_location is None"
                )
            encoded_data = (
                self._get_bytes(string_location, size=size, code_page=code_page) or b""
            )

        if not isinstance(encoded_data, (bytes, list)):
            raise TypeError(
                "encoded_data must be type 'bytes', not '{!r}'".format(
                    type(encoded_data)
                )
            )

        self.encoded_data = encoded_data

    @classmethod
    def factory(
        cls,
        string_location,
        string_reference,
        size=None,
        offset=None,
        key=None,
        encoded_data=None,
        code_page=None,
        dest=None,
    ):
        """
        Factory function to generate an EncodedString or EncodedStackString based on type.

        :param string_location:
            Data segment pointer for static strings or stack pointer for stack strings.
        :param string_reference:
            The location the string is referenced from.
            This is required to pull the stack frame when string_location is a stack pointer.
        :param size: Size of the encoded data.
            When set encoded_data is populated using this and string_location.
        :param offset: Used when there is an offset based accessing scheme.
        :param key: Used when there is a key that can vary by string.
        :param encoded_data: encoded/encrypted data, if not provided data will be retrieved from IDA.
        :param code_page: known encoding page used to decode data to unicode (after data is decrypted)
            (code page is dynamically determined if not provided)
        :param dest: Location of decrypted data (if different from string_location)
        """
        if idc.is_loaded(string_location):
            return EncodedString(
                string_location,
                string_reference,
                size=size,
                offset=offset,
                key=key,
                encoded_data=encoded_data,
                code_page=code_page,
                dest=dest,
            )

        # otherwise assume string_location is a pointer within the stack
        # (using function_tracing's CPU emulator) and create an EncodedStackString object.
        stack = idc.get_func_attr(string_reference, idc.FUNCATTR_FRAME)
        # FIXME: This method isn't always super accurate because... IDA
        stack_offset = (
            string_location
            + function_tracing.RSP_OFFSET
            + idc.get_func_attr(string_reference, idc.FUNCATTR_FRSIZE)
            - function_tracing.STACK_BASE
        )
        if stack_offset < 0:
            logger.warning(
                "Ignoring negative stack offset {:#x} pulled from 0x{:X}".format(
                    stack_offset, string_location
                )
            )
            stack_offset = None

        return EncodedStackString(
            encoded_data,
            frame_id=stack,
            stack_offset=stack_offset,
            string_reference=string_reference,
            offset=offset,
            key=key,
            code_page=code_page,
            dest=dest,
        )

    def _compare_key(self):
        # Sort by where it was found then where it is referenced
        return (
            self.string_location or -1,
            self.string_reference or -1,
            self.decoded_data or b"",
        )

    def __hash__(self):
        return hash(self._compare_key())

    def __eq__(self, other):
        return self._compare_key() == other._compare_key()

    def __lt__(self, other):
        return self._compare_key() < other._compare_key()

    def report(self):
        """
        Generates a text report of the EncodedString object.

        :return str: Unicode string containing text report.
        """
        # We have to return repr as a unicode because of the decoded string.
        text = u""
        if self.string_location is not None:
            text += u"EA:  0x{:08X}\n".format(self.string_location)
        if self.string_reference is not None:
            text += u"Ref: 0x{:08X}\n".format(self.string_reference)
        if self.dest is not None:
            text += u"Dest: 0x{:08X}\n".format(self.dest)
        if self.offset is not None:
            text += u"Offset: 0x{:08X}\n".format(self.offset)
        if self.encoded_data:
            text += u"Raw Enc: {!r}\n".format(self.encoded_data)
        if self.decoded_data:
            text += u"Raw Dec: {!r}\n".format(self.decoded_data)
            dec_string = str(self)
            if self.code_page:
                text += u"Detected Code Page: {}\n".format(self.code_page)
            text += u"Dec: {}".format(dec_string)
        return text

    def __bytes__(self):
        return self.decoded_data

    def __str__(self):
        return self._decode_unknown_charset()

    def __unicode__(self):
        return self._decode_unknown_charset()

    def __repr__(self):
        encoded_data = self.encoded_data or b""
        if len(encoded_data) > 30:
            encoded_data = encoded_data[:30] + b" ..."
        return "<{!r} at 0x{:08x}>".format(
            encoded_data, self.string_location or self.string_reference
        )

    def __len__(self):
        return len(self.encoded_data)

    @property
    def xrefs_to(self):
        if self._xrefs_to is None:
            self._xrefs_to = [ref.frm for ref in idautils.XrefsTo(self.string_location)]
        return self._xrefs_to

    @property
    def display_name(self):
        """Returns an IDA friendly, printable name for the decoded string."""
        return str(self).encode(DISPLAY_CODE, "replace").decode(DISPLAY_CODE)

    def rename(self, name=None):
        """
        Renames (and comments) the string variable in IDA.

        :param str name: New name to given encoded string. (defaults to decoded_string)
        """
        name = name or self.display_name
        if not name:
            logger.warning(
                "Unable to rename encoded string due to no decoded string: {!r}".format(
                    self
                )
            )
            return

        # Add comment
        comment = '"{}"'.format(name[: self._MAX_COMMENT_LENGTH])
        if len(name) > self._MAX_COMMENT_LENGTH:
            comment += " (truncated)"
        if self.string_location is not None:
            idc.set_cmt(self.string_location, comment, 1)
        if self.string_reference is not None:
            idc.set_cmt(self.string_reference, comment, 1)
        if self.dest is not None:
            idc.set_cmt(self.dest, comment, 1)

        # Set variable name
        # To follow IDA conventions, the decoded string itself will be prefixed with 'a'
        # If a destination was provided, don't include any prefix, because this
        # is usually for resolved API functions.
        name = name[: self._MAX_NAME_LENGTH]
        if self.string_location is not None:
            ida_name.force_name(self.string_location, "a" + name)
        if self.dest is not None:
            # IDA will sometimes type API function pointers based on code analysis, and it's generic.  Unsetting the
            # type before the rename will cause IDA to properly type API functions if the type is known from the
            # loaded type libraries.
            idc.SetType(self.dest, '')
            ida_name.force_name(self.dest, name)

    def patch(self, fill_char=b'\x00', define=True):
        """
        Patches the original encoded string with the decoded string.

        :param str fill_char:
            Character to use to fill left over space if decoded data
            is shorter than its encoded data.
            Set to None leaving the original data.
        :param bool define: Whether to define the string after patching.
        """
        if not self.decoded_data or self.string_location is None:
            return
        decoded_data = self.decoded_data
        if fill_char:
            decoded_data += fill_char * (len(self.encoded_data) - len(decoded_data))
        try:
            idaapi.patch_bytes(self.start_ea, decoded_data)
            if define:
                self.define()
        except TypeError:
            logger.debug(
                "String type for decoded string from location 0x{:08x}.".format(
                    self.start_ea
                )
            )
        finally:
            return self

    # TODO: Can this just be embedded in patch()?
    def define(self):
        """
        Defines the string in the IDB.
        """
        try:
            idc.del_items(self.start_ea, idc.DELIT_SIMPLE, len(self.decoded_data))
            idaapi.create_strlit(
                self.start_ea, len(self.decoded_data), self.string_type
            )
        except Exception as e:
            logger.warning(
                "Unable to define string at 0x{:0X}: {}".format(self.start_ea, e)
            )

    def publish(self, rename=True, patch=True, fill_char=b'\x00', define=True):
        """
        - Saves encoded string to external kordesii Reporter
        - Prints a report about the string to the console
        - renames and patches the IDB with decoded data

        :param rename: Whether to rename the string in the IDB.
        :param patch: Whether to patch the string with the decoded variant in the IDB.
        :param str fill_char:
            Character to use to fill left over space if decoded data
            is shorter than its encoded data.
            Set to None leaving the original data.
        :param bool define: Whether to define the string after patching.
        """
        if not self.decoded_data:
            logger.warning(
                "Unable to publish string {!r}. Missing decoded_data.".format(self)
            )
            return

        # FIXME: Even though we strip nulls in __unicode__(), there still seems to be some strings
        # with null characters seeping through.
        kordesii.append_string(str(self).rstrip(u"\x00"))

        print("\n")
        display = self.report()
        print(display)

        if rename:
            self.rename()
        if patch:
            self.patch(fill_char=fill_char, define=define)

    @property
    def start_ea(self):
        if self.string_location in (None, idc.BADADDR):
            return -1
        else:
            return self.string_location + (self.offset or 0)

    @property
    def end_ea(self):
        start_ea = self.start_ea
        if start_ea is None:
            return -1
        else:
            return start_ea + len(self.decoded_data)

    @property
    def string_type(self):
        if not self.decoded_data:
            return None
        else:
            return (
                idc.STRTYPE_C_16
                if isinstance(self.decoded_data, str)
                else idc.STRTYPE_C
            )

    def _get_bytes(self, location, size=None, code_page=None):
        """
        Extracts bytes from given location.

        :param location: Location to pull bytes
        :param size: Number of bytes to pull (determines size by looking for terminator if not provided)
        :param code_page: Known code_page used to determine terminator.
        :return: bytes or None
        """
        # Determine size by looking for terminator.
        # (Use provided encoding to determine terminator width.)
        if size is None:
            width = 1
            if code_page:
                if "16" in code_page:
                    width = 2
                elif "32" in code_page:
                    width = 4
            end_location = idc.find_binary(location, idc.SEARCH_DOWN, "00 " * width)
            if end_location == idc.BADADDR:
                logger.warning("Failed to extract bytes from 0x{:08X}".format(location))
                return None

            size = end_location - location
            while size % width:  # ensure unicode strings are a valid length
                size += 1

        # Pull size amount of bytes from IDA.
        data = idc.get_bytes(location, size)
        if data is None:
            logger.warning(
                "Failed to extract {} bytes from 0x{:08X}".format(size, location)
            )
        return data

    def _num_raw_bytes(self, string):
        """
        Returns the number of raw bytes found in the given unicode string
        """
        count = 0
        for char in string:
            char = char.encode("unicode-escape")
            count += char.startswith(b"\\x") + char.startswith(b"\\u") * 2
        return count

    def _decode_unknown_charset(self):
        """
        Returns a decoded string using the best guess codec.
        """
        if not self.decoded_data:
            return u""

        # First see if the decoder already gave us unicode.
        if isinstance(self.decoded_data, str):
            return self.decoded_data

        # If a code page was set (either by us or the decoder) use that.
        # If the code page doesn't work... move on.
        if self.code_page:
            try:
                return self.decoded_data.decode(self.code_page)
            except UnicodeDecodeError:
                pass

        # TODO: Use chardet if they ever support utf16 without BOM.
        best_score = len(self.decoded_data)
        best_code_page = None
        best_output = None
        for code_page in CODE_PAGES:
            try:
                output = self.decoded_data.decode(code_page).rstrip(u"\x00")
            except UnicodeDecodeError:
                # If it's UTF we may need to strip away some null characters before decoding.
                if code_page in ("utf-16-le", "utf-16-be", "utf-32-le", "utf-32-be"):
                    decoded_data = self.decoded_data
                    while decoded_data and decoded_data[-1] == 0:
                        try:
                            decoded_data = decoded_data[:-1]
                            output = decoded_data.decode(code_page).rstrip(u"\x00")
                        except UnicodeDecodeError:
                            continue
                        break  # successfully decoded
                    else:
                        continue
                # otherwise the code page isn't correct.
                else:
                    continue

            score = self._num_raw_bytes(output)
            if not best_output or score < best_score:
                best_score = score
                best_output = output
                best_code_page = code_page

        if best_output:
            self.code_page = best_code_page
            return best_output

        return u""

    def split(self, define=False) -> List["EncodedString"]:
        """
        Splits up EncodedString object into a list of multiple EncodedString objects split up by
        null terminators, which can contain utf-8 or utf-16 encoded strings.

        NOTE: This function assumes one-to-one decryption has occurred.
            If not, you can call find_string_data() manually.

        The original EncodedString is not modified.

        :param define: Whether to define new strings in IDB.
        """
        new_strings = []
        for offset, string_data, encoding in find_string_data(self.decoded_data):
            new_string = copy.copy(self)
            new_string.encoded_data = self.encoded_data[offset: offset + len(string_data)]
            new_string.decoded_data = string_data
            new_string.offset = offset
            if define:
                new_string.define()
            new_strings.append(new_string)
        return new_strings


@serializable_class(skip_attrs=["xrefs_to"])
class EncodedStackString(EncodedString):
    """
    Variant of EncodedString that represents a string built from the stack.
    """

    def __init__(
        self,
        encoded_data,
        frame_id=None,
        stack_offset=None,
        memory_ptr=None,
        string_reference=None,
        offset=None,
        key=None,
        code_page=None,
        dest=None,
    ):
        """
        Instantiate an EncodedStackString object.

        :param encoded_data:
            encoded/encrypted data of string
            (This must be provided because stack strings don't have a string_location attribute.)
        :param frame_id: The id of the IDA frame containing the stack this string comes from.
        :param stack_offset: The offset within the IDA frame containing this string.
        :param memory_ptr: Optional extra argument to keep track of pointer to string in memory
            when using function_tracing.
        :param string_reference:
            The location the string is referenced from.
            This is required to pull the stack frame when string_location is a stack pointer.
        :param offset: Used when there is an offset based accessing scheme.
        :param key: Used when there is a key that can vary by string.
        :param code_page: known encoding page used to decode data to unicode (after data is decrypted)
            (code page is dynamically determined if not provided)
        :param dest: Location of decrypted data
        """
        super(EncodedStackString, self).__init__(
            None,
            string_reference=string_reference,
            offset=offset,
            key=key,
            encoded_data=encoded_data,
            code_page=code_page,
            dest=dest,
        )
        # Frame ID and Stack Offset are optional because it's not always easy to calculate this.
        self.frame_id = frame_id
        self.stack_offset = stack_offset
        # TODO: Remove memory_ptr, it should be manually added by the decoder after initialization
        # if they want to use it.
        # Optional extra argument to keep track of pointer to string in memory when using function_tracing.
        self.memory_ptr = memory_ptr

    def _compare_key(self):
        # Sort by where it was found and then by offset within stack.
        return (
            self.string_reference,
            self.frame_id,
            self.stack_offset,
            self.decoded_data,
        )

    def report(self):
        """
        General display format.

        :return unicode:
            A unicode string with the string's frame ID and stack offset,
            reference EA (where applicable), and the decoded value (where applicable).
        """
        text = u""
        if self.frame_id:
            text += u"Frame ID: 0x%X\n" % self.frame_id
        if self.stack_offset:
            text += u"Stack Offset: 0x%X\n" % self.stack_offset
        text += super(EncodedStackString, self).report()
        return text

    @property
    def xrefs_to(self):
        """
        Retrieves the xrefs to the stack variable.

        NOTE: This code is very SWIGGY because IDA did not properly expose this functionality.

        :raises ValueError: if frame_id, stack_offset, or string_reference was not provided.
            This is needed to determine what function to use.
        """
        if self._xrefs_to is None:
            if not self.string_reference:
                raise ValueError("Unable to get xrefs without string_reference.")
            if not (self.frame_id and self.stack_offset):
                raise ValueError(
                    "Unable to get xrefs without frame_id and stack_offset"
                )
            xrefs = idaapi.xreflist_t()
            frame = idaapi.get_frame(self.frame_id)
            func = idaapi.get_func(self.string_reference)
            member = idaapi.get_member(frame, self.stack_offset)
            idaapi.build_stkvar_xrefs(xrefs, func, member)
            self._xrefs_to = [ref.ea for ref in xrefs]
        return self._xrefs_to

    def rename(self, name=None):
        """
        Renames (and comments) the string variable in IDA.

        :param str name: New name to given encoded string. (defaults to decoded_string)
        """
        name = name or self.display_name
        if not name:
            logger.warning(
                "Unable to rename encoded string due to no decoded string: {!r}".format(
                    self
                )
            )

        # Set name and comment in stack variable.
        comment = '"{}"'.format(name[: self._MAX_COMMENT_LENGTH])
        if len(name) > self._MAX_COMMENT_LENGTH:
            comment += " (truncated)"
        if self.frame_id and self.stack_offset:
            idc.set_member_cmt(self.frame_id, self.stack_offset, comment, repeatable=1)
            var_name = re.sub(
                "[^_$?@0-9A-Za-z]", "_", name[: self._MAX_NAME_LENGTH]
            )  # Replace invalid characters
            if not var_name:
                raise ValueError("Unable to calculate var_name for : {!r}".format(self))
            var_name = "a" + var_name.capitalize()
            idc.set_member_name(self.frame_id, self.stack_offset, var_name)

        if self.dest is not None:
            ida_name.force_name(self.dest, name)

        # Add a comment where the string is being used.
        if self.string_reference:
            idc.set_cmt(self.string_reference, comment, 1)

    def patch(self, fill_char=None, define=True):
        """Does nothing, patching is not a thing for stack strings."""
        return self



"""
Interface for operand management in x86.
"""

import logging
from typing import Optional

import idc

from kordesii.utils.function_tracing import utils
from kordesii.utils.function_tracing.operands import Operand
from kordesii.utils.function_tracing.x86_64 import ida_intel

logger = logging.getLogger(__name__)


class x86_64Operand(Operand):

    def _calc_displacement(self):
        """
        Calculate the displacement offset of the operand's text.

        e.g:
            word ptr [rdi+rbx]

        :return int: calculated value
        """
        addr = self.base + self.index * self.scale + self.offset
        logger.debug(
            "Calculating operand: %s -> 0x%X + 0x%X*0x%X %s 0x%X = 0x%X" % (
                self.text,
                self.base,
                self.index,
                self.scale,
                "-" if self.offset < 0 else "+",
                abs(self.offset),
                addr
            )
        )
        if addr < 0:
            logger.debug("Address is negative, resorting to address of 0.")
            addr = 0

        return addr

    @property
    def base(self) -> Optional[int]:
        """
        The value of the base register if operand is a displacement.

        e.g.
            [ebp+ecx*2+var_8] -> ebp
        """
        if not self.has_phrase:
            return None
        base_reg = ida_intel.x86_base_reg(self._insn, self._op)
        value = self._cpu_context.registers[utils.reg2str(base_reg)]
        return utils.signed(value)

    @property
    def scale(self) -> Optional[int]:
        """
        The scaling factor of the index if operand is a displacement.

        e.g.
            [ebp+ecx*2+var_8] -> 2
        """
        if not self.has_phrase:
            return None
        return 1 << ida_intel.sib_scale(self._op)

    @property
    def index(self) -> Optional[int]:
        """
        The value of the index register if operand is a displacement.

        e.g.
            [ebp+ecx*2+var_8] -> ecx
        """
        if not self.has_phrase:
            return None
        index_reg = ida_intel.x86_index_reg(self._insn, self._op)
        if index_reg == -1:
            return 0
        index_reg = utils.reg2str(index_reg)
        value = self._cpu_context.registers[index_reg]
        return utils.signed(value)

    @property
    def offset(self) -> Optional[int]:
        """
        The offset value if the operand is a displacement.

        e.g.
            [ebp+ecx*2+var_8] -> var_8
        """
        if not self.has_phrase:
            return None
        return utils.signed(self._op.addr)

    @property
    def addr(self) -> Optional[int]:
        """
        Retrieves the referenced memory address of the operand.

        :return int: Memory address or None if operand is not a memory reference.
        """
        addr = None
        if self.has_phrase:
            # These need to be handled in the same way even if they don't contain the same types of data...
            addr = self._calc_displacement()
            # Before returning, record the stack variable that we have encountered.
            self._record_stack_variable(addr)
        elif self.type == idc.o_mem:
            addr = self._op.addr
            # Record the global variable before we return.
            self._cpu_context.variables.add(addr, reference=self.ip)
        return addr

    @property
    def base_addr(self):
        """
        Retrieves the referenced memory address of the operand minus any indexing that
        has occurred.

        This is useful for pulling out the un-offseted address within a loop.
        e.g. "movzx   edx, [ebp+ecx*2+var_8]"
        where ecx is the loop index starting at a non-zero value.

        :return int: Memory address or None if operand is not a memory reference.
        """
        addr = self.addr
        if addr is None:
            return None
        if self.has_phrase:
            addr -= self.index * self.scale
        return addr

    @property
    def value(self):
        # TODO: Determine if this is still necessary.
        # FS, GS (at least) registers are identified as memory addresses.  We need to identify them as registers
        # and handle them as such
        if self.type == idc.o_mem:
            if "fs" in self.text:
                return self._cpu_context.registers.fs
            elif "gs" in self.text:
                return self._cpu_context.registers.gs

        return super().value

    @value.setter
    def value(self, value):
        try:
            logger.debug("0x%X -> %s", value, self.text)
        except TypeError:
            logger.debug("%r -> %s", value, self.text)

        # TODO: Determine if this is still necessary.
        # FS, GS (at least) registers are identified as memory addresses.  We need to identify them as registers
        # and handle them as such
        if self.type == idc.o_mem:
            if "fs" in self.text:
                self._cpu_context.registers.fs = value
                return
            elif "gs" in self.text:
                self._cpu_context.registers.gs = value
                return

        # On 64-bit, the destination register must be set to 0 first (per documentation)
        # TODO: Check if this happens regardless of the source size
        if self.is_register and  idc.__EA64__ and self.width == 4:  # Only do this for 32-bit setting
            reg64 = utils.convert_reg(self.text, 8)
            self._cpu_context.registers[reg64] = 0

        super(x86_64Operand, self.__class__).value.__set__(self, value)



"""
Interface for instruction management.
"""
import logging

import idc

from kordesii.utils.function_tracing.instruction import Instruction
from kordesii.utils.function_tracing.x86_64.operands import x86_64Operand

logger = logging.getLogger(__name__)


class x86_64Instruction(Instruction):

    _operand_class = x86_64Operand

    def _execute(self):
        # Before executing, determine if a rep* instruction and add termination condition.
        term_condition = None
        if idc.get_wide_byte(self.ip) in (0xF2, 0xF3):
            text = self.text  # IDA pro never has operands for rep opcodes, we need to look at the text.
            if text.startswith("rep "):
                term_condition = lambda ctx: ctx.registers.ecx == 0
                term_condition.unconditional = True
            elif text.startswith(("repe ", "repz ")):
                term_condition = lambda ctx: ctx.registers.ecx == 0 or ctx.registers.zf == 0
                term_condition.unconditional = False
            elif text.startswith(("repne ", "repnz ")):
                term_condition = lambda ctx: ctx.registers.ecx == 0 or ctx.registers.zf == 1
                term_condition.unconditional = False

        # Execute like normal if not a rep instruciton.
        if not term_condition:
            super()._execute()
            return

        # Skip if user disabled all rep instruction.
        if self._cpu_context.emulator.disabled_rep:
            logger.debug("Ignoring rep instruction at 0x%X: DISABLED.", self.ip)
            return

        # As a safety measure, don't allow rep instructions to surpass
        # our max memory read limit.
        # Only do this check if the terminating condition is unconditional, otherwise
        # this number usually big because it expects zf to be toggled.
        if (
                term_condition.unconditional
                and self._cpu_context.registers.ecx > self._cpu_context.memory.MAX_MEM_READ
        ):
            logger.warning(
                "Emulation attempted to read %s instruction %d times. "
                "Ignoring instruction.", self.mnem, self._cpu_context.registers.ecx
            )
            return

        # Execute instruction ecx number of times or when terminal condition applies.
        logger.debug("Emulating %s instruction %d times.", self.mnem, self._cpu_context.registers.ecx)
        count = 0
        while not term_condition(self._cpu_context):
            super()._execute()
            self._cpu_context.registers.ecx -= 1
            # Stop if we are iterating too much.
            count += 1
            if count > self._cpu_context.memory.MAX_MEM_READ:
                logger.warning("Looped too many times, exiting prematurely.")
                break
